"""
Web handlers for the FractalServer.
"""
import json

import tornado.web
from pydantic import ValidationError
from qcelemental.util import deserialize, serialize

from .interface.models.rest_models import rest_model
from .storage_sockets.storage_utils import add_metadata_template

_valid_encodings = {
    "application/json": "json",
    "application/json-ext": "json-ext",
    "application/msgpack-ext": "msgpack-ext",
}


class APIHandler(tornado.web.RequestHandler):
    """
    A requests handler for API calls.
    """

    # Admin authentication required by default
    _required_auth = "admin"
    _logging_param_counts = {}

    def initialize(self, **objects):
        """
        Initializes the request to JSON, adds objects, and logging.
        """

        self.content_type = "Not Provided"
        try:
            # default to "application/json"
            self.content_type = self.request.headers.get("Content-Type", "application/json")
            self.encoding = _valid_encodings[self.content_type]
        except KeyError:
            raise tornado.web.HTTPError(
                status_code=401, reason=f"Did not understand 'Content-Type': {self.content_type}"
            )

        # Always reply in the format sent
        self.set_header("Content-Type", self.content_type)

        self.objects = objects
        self.storage = self.objects["storage_socket"]
        self.logger = objects["logger"]
        self.api_logger = objects["api_logger"]
        self.view_handler = objects["view_handler"]
        self.username = None

    def prepare(self):
        if self._required_auth:
            self.authenticate(self._required_auth)

        try:
            if (self.encoding == "json") and isinstance(self.request.body, bytes):
                blob = self.request.body.decode()
            else:
                blob = self.request.body

            if blob:
                self.data = deserialize(blob, self.encoding)
            else:
                self.data = None
        except:
            raise tornado.web.HTTPError(status_code=401, reason="Could not deserialize body.")

    def on_finish(self):

        exclude_uris = ["/task_queue", "/service_queue", "/queue_manager"]

        # No associated data, so skip all of this
        # (maybe caused by not using portal or not using the REST API correctly?)
        if self.data is None:
            return

        if self.api_logger and self.request.method == "GET" and self.request.uri not in exclude_uris:

            extra_params = self.data.copy()
            if self._logging_param_counts:
                for key in self._logging_param_counts:
                    if extra_params["data"].get(key, None):
                        extra_params["data"][key] = len(extra_params["data"][key])

            if "data" in extra_params:
                extra_params["data"] = {k: v for k, v in extra_params["data"].items() if v is not None}

            extra_params = json.dumps(extra_params)

            log = self.api_logger.get_api_access_log(request=self.request, extra_params=extra_params)
            self.storage.save_access(log)

        # self.logger.info('Done saving API access to the database')

    def authenticate(self, permission):
        """Authenticates request with a given permission setting.

        Parameters
        ----------
        permission : str
            The required permission ["read", "write", "compute", "admin"]

        """
        if "Authorization" in self.request.headers:

            data = json.loads(self.request.headers["Authorization"])
            username = data["username"]
            password = data["password"]
        else:
            username = None
            password = None

        self.username = username

        verified, msg = self.objects["storage_socket"].verify_user(username, password, permission)
        if verified is False:
            raise tornado.web.HTTPError(status_code=401, reason=msg)

    def parse_bodymodel(self, model):

        try:
            return model(**self.data)
        except ValidationError:
            raise tornado.web.HTTPError(status_code=401, reason="Invalid REST")

    def write(self, data):
        if not isinstance(data, (str, bytes)):
            data = serialize(data, self.encoding)

        return super().write(data)


class InformationHandler(APIHandler):
    """
    A handler that returns public server information.
    """

    _required_auth = "read"

    def get(self):
        """"""

        self.logger.info("GET: Information")

        self.write(self.objects["public_information"])


class KVStoreHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    _required_auth = "read"
    _logging_param_counts = {"id"}

    def get(self):
        """

        Experimental documentation, need to find a decent format.

        Request:
            "data" - A list of key requests

        Returns:
            "meta" - Metadata associated with the query
                - "errors" - A list of errors in (index, error_id) format.
                - "n_found" - The number of molecule found.
                - "success" - If the query was successful or not.
                - "error_description" - A string based description of the error or False
                - "missing" - A list of keys that were not found.
            "data" - A dictionary of {key : value} dictionary of the results

        """

        body_model, response_model = rest_model("kvstore", "get")
        body = self.parse_bodymodel(body_model)

        ret = self.storage.get_kvstore(body.data.id)
        ret = response_model(**ret)

        self.logger.info("GET: KVStore - {} pulls.".format(len(ret.data)))
        self.write(ret)


class WavefunctionStoreHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    _required_auth = "read"
    _logging_param_counts = {"id"}

    def get(self):

        body_model, response_model = rest_model("wavefunctionstore", "get")
        body = self.parse_bodymodel(body_model)

        ret = self.storage.get_wavefunction_store(body.data.id, include=body.meta.include)
        if len(ret["data"]):
            ret["data"] = ret["data"][0]
        ret = response_model(**ret)

        self.logger.info("GET: WavefunctionStore - 1 pull.")
        self.write(ret)


class MoleculeHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    _required_auth = "read"
    _logging_param_counts = {"id"}

    def get(self):
        """

        Experimental documentation, need to find a decent format.

        Request:
            "meta" - Overall options to the Molecule pull request
                - "index" - What kind of index used to find the data ("id", "molecule_hash", "molecular_formula")
            "data" - A dictionary of {key : index} requests

        Returns:
            "meta" - Metadata associated with the query
                - "errors" - A list of errors in (index, error_id) format.
                - "n_found" - The number of molecule found.
                - "success" - If the query was successful or not.
                - "error_description" - A string based description of the error or False
                - "missing" - A list of keys that were not found.
            "data" - A dictionary of {key : molecule JSON} results

        """

        body_model, response_model = rest_model("molecule", "get")
        body = self.parse_bodymodel(body_model)

        molecules = self.storage.get_molecules(**{**body.data.dict(), **body.meta.dict()})
        ret = response_model(**molecules)

        self.logger.info("GET: Molecule - {} pulls.".format(len(ret.data)))
        self.write(ret)

    def post(self):
        """
            Experimental documentation, need to find a decent format.

        Request:
            "meta" - Overall options to the Molecule pull request
                - No current options
            "data" - A dictionary of {key : molecule JSON} requests

        Returns:
            "meta" - Metadata associated with the query
                - "errors" - A list of errors in (index, error_id) format.
                - "n_inserted" - The number of molecule inserted.
                - "success" - If the query was successful or not.
                - "error_description" - A string based description of the error or False
                - "duplicates" - A list of keys that were already inserted.
            "data" - A dictionary of {key : id} results
        """

        self.authenticate("write")

        body_model, response_model = rest_model("molecule", "post")
        body = self.parse_bodymodel(body_model)

        ret = self.storage.add_molecules(body.data)
        response = response_model(**ret)

        self.logger.info("POST: Molecule - {} inserted.".format(response.meta.n_inserted))
        self.write(response)


class KeywordHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    _required_auth = "read"
    _logging_param_counts = {"id"}

    def get(self):

        body_model, response_model = rest_model("keyword", "get")
        body = self.parse_bodymodel(body_model)

        ret = self.storage.get_keywords(**{**body.data.dict(), **body.meta.dict()}, with_ids=False)
        response = response_model(**ret)

        self.logger.info("GET: Keywords - {} pulls.".format(len(response.data)))
        self.write(response)

    def post(self):
        self.authenticate("write")

        body_model, response_model = rest_model("keyword", "post")
        body = self.parse_bodymodel(body_model)

        ret = self.storage.add_keywords(body.data)
        response = response_model(**ret)

        self.logger.info("POST: Keywords - {} inserted.".format(response.meta.n_inserted))
        self.write(response)


class CollectionHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    _required_auth = "read"

    def get(self, collection_id=None, view_function=None):

        # List collections
        if (collection_id is None) and (view_function is None):
            body_model, response_model = rest_model("collection", "get")
            body = self.parse_bodymodel(body_model)

            cols = self.storage.get_collections(
                **body.data.dict(), include=body.meta.include, exclude=body.meta.exclude
            )
            response = response_model(**cols)

            self.logger.info("GET: Collections - {} pulls.".format(len(response.data)))
            self.write(response)
            return

        # Get specific collection
        elif (collection_id is not None) and (view_function is None):
            body_model, response_model = rest_model("collection", "get")

            body = self.parse_bodymodel(body_model)
            cols = self.storage.get_collections(
                **body.data.dict(), col_id=int(collection_id), include=body.meta.include, exclude=body.meta.exclude
            )
            response = response_model(**cols)

            self.logger.info("GET: Collections - {} pulls.".format(len(response.data)))
            self.write(response)
            return

        # View-backed function on collection
        elif (collection_id is not None) and (view_function is not None):
            body_model, response_model = rest_model(f"collection/{collection_id}/{view_function}", "get")
            body = self.parse_bodymodel(body_model)
            if self.view_handler is None:
                meta = {
                    "success": False,
                    "error_description": "Server does not support collection views.",
                    "errors": [],
                    "msgpacked_cols": [],
                }
                self.write(response_model(meta=meta, data=None))
                self.logger.info("GET: Collections - view request made, but server does not have a view_handler.")
                return

            result = self.view_handler.handle_request(collection_id, view_function, body.data.dict())
            response = response_model(**result)

            self.logger.info(f"GET: Collections - {collection_id} view {view_function} pulls.")
            self.write(response)
            return

        # Unreachable?
        else:
            body_model, response_model = rest_model("collection", "get")
            meta = add_metadata_template()
            meta["success"] = False
            meta["error_description"] = "GET request for view with no collection ID not understood."
            self.write(response_model(meta=meta, data=None))
            self.logger.info(
                "GET: Collections - collection id is None, but view function is not None (should be unreachable)."
            )
            return

    def post(self, collection_id=None, view_function=None):
        self.authenticate("write")

        body_model, response_model = rest_model("collection", "post")
        body = self.parse_bodymodel(body_model)

        # POST requests not supported for anything other than "/collection"
        if collection_id is not None or view_function is not None:
            meta = add_metadata_template()
            meta["success"] = False
            meta["error_description"] = "POST requests not supported for sub-resources of /collection"
            self.write(response_model(meta=meta, data=None))
            self.logger.info("POST: Collections - Access attempted on subresource.")
            return

        ret = self.storage.add_collection(body.data.dict(), overwrite=body.meta.overwrite)
        response = response_model(**ret)

        self.logger.info("POST: Collections - {} inserted.".format(response.meta.n_inserted))
        self.write(response)

    def delete(self, collection_id, _):
        self.authenticate("write")

        body_model, response_model = rest_model(f"collection/{collection_id}", "delete")
        ret = self.storage.del_collection(col_id=collection_id)
        if ret == 0:
            self.logger.info(f"DELETE: Collections - Attempted to delete non-existent collection {collection_id}.")
            raise tornado.web.HTTPError(status_code=404, reason=f"Collection {collection_id} does not exist.")
        else:
            self.write(response_model(meta={"success": True, "errors": [], "error_description": False}))
            self.logger.info(f"DELETE: Collections - Deleted collection {collection_id}.")


class ResultHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    _required_auth = "read"
    _logging_param_counts = {"id", "molecule"}

    def get(self):

        body_model, response_model = rest_model("result", "get")
        body = self.parse_bodymodel(body_model)

        ret = self.storage.get_results(**{**body.data.dict(), **body.meta.dict()})
        result = response_model(**ret)

        self.logger.info("GET: Results - {} pulls.".format(len(result.data)))
        self.write(result)


class ProcedureHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    _required_auth = "read"
    _logging_param_counts = {"id"}

    def get(self, query_type="get"):

        body_model, response_model = rest_model("procedure", query_type)
        body = self.parse_bodymodel(body_model)

        try:
            if query_type == "get":
                ret = self.storage.get_procedures(**{**body.data.dict(), **body.meta.dict()})
            else:  # all other queries, like 'best_opt_results'
                ret = self.storage.custom_query("procedure", query_type, **{**body.data.dict(), **body.meta.dict()})
        except KeyError as e:
            raise tornado.web.HTTPError(status_code=401, reason=str(e))

        response = response_model(**ret)

        self.logger.info("GET: Procedures - {} pulls.".format(len(response.data)))
        self.write(response)


class OptimizationHandler(APIHandler):
    """
    A handler to push and get molecules.
    """

    _required_auth = "read"
    _logging_param_counts = {"id"}

    def get(self, query_type="get"):

        body_model, response_model = rest_model(f"optimization/{query_type}", "get")
        body = self.parse_bodymodel(body_model)

        try:
            if query_type == "get":
                ret = self.storage.get_procedures(**{**body.data.dict(), **body.meta.dict()})
            else:  # all other queries, like 'best_opt_results'
                ret = self.storage.custom_query("optimization", query_type, **{**body.data.dict(), **body.meta.dict()})
        except KeyError as e:
            raise tornado.web.HTTPError(status_code=401, reason=str(e))

        response = response_model(**ret)

        self.logger.info("GET: Optimization ({}) - {} pulls.".format(query_type, len(response.data)))
        self.write(response)



"""Msgpack Remaining Phase 1

Revision ID: da7c6f141bcb
Revises: 05ceea11b78a
Create Date: 2019-08-12 10:12:46.478628

"""
from alembic import op
import sqlalchemy as sa

import os
import sys

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))
from migration_helpers import msgpack_migrations
from qcelemental.util import msgpackext_dumps, msgpackext_loads

# revision identifiers, used by Alembic.
revision = "da7c6f141bcb"
down_revision = "05ceea11b78a"
branch_labels = None
depends_on = None

block_size = 100


def transformer(old_data):

    extras = old_data["extras"]
    extras.pop("_qcfractal_tags", None)  # cleanup old tags

    return {"extras_": msgpackext_dumps(extras)}


def upgrade():

    ## Task Queue
    table_name = "task_queue"
    update_columns = {"spec"}

    def transformer(old_data):

        spec = old_data["spec"]

        return {"spec_": msgpackext_dumps(spec)}

    msgpack_migrations.json_to_msgpack_table(table_name, block_size, update_columns, transformer)

    ## Service Queue
    table_name = "service_queue"
    update_columns = {"extra"}

    def transformer(old_data):

        spec = old_data["extra"]

        return {"extra_": msgpackext_dumps(spec)}

    msgpack_migrations.json_to_msgpack_table(table_name, block_size, update_columns, transformer, {})


def downgrade():

    ## Task Queue
    table_name = "task_queue"
    update_columns = {"spec"}
    msgpack_migrations.json_to_msgpack_table_dropcols(table_name, block_size, update_columns)

    ## Service Queue
    table_name = "service_queue"
    update_columns = {"extra"}
    msgpack_migrations.json_to_msgpack_table_dropcols(table_name, block_size, update_columns)



import hashlib
import json
from typing import Any, Dict, Optional

import numpy as np

json_encoders = {np.ndarray: lambda v: v.flatten().tolist()}


def prepare_basis(basis: Optional[str]) -> Optional[str]:
    """
    Prepares a basis set string
    """
    if basis is None:
        return basis

    if basis == "":
        return None

    if basis == "null":
        return None

    return basis.lower()


def recursive_normalizer(value: Any, **kwargs: Dict[str, Any]) -> Any:
    """
    Prepare a structure for hashing by lowercasing all values and round all floats
    """
    digits = kwargs.get("digits", 10)
    lowercase = kwargs.get("lowercase", True)

    if isinstance(value, (int, type(None))):
        pass

    elif isinstance(value, str):
        if lowercase:
            value = value.lower()

    elif isinstance(value, list):
        value = [recursive_normalizer(x, **kwargs) for x in value]

    elif isinstance(value, tuple):
        value = tuple(recursive_normalizer(x, **kwargs) for x in value)

    elif isinstance(value, dict):
        ret = {}
        for k, v in value.items():
            if lowercase:
                k = k.lower()
            ret[k] = recursive_normalizer(v, **kwargs)
        value = ret

    elif isinstance(value, np.ndarray):
        if digits:
            # Round array
            value = np.around(value, digits)
            # Flip zeros
            value[np.abs(value) < 5 ** (-(digits + 1))] = 0

    elif isinstance(value, float):
        if digits:
            value = round(value, digits)
            if value == -0.0:
                value = 0
            if value == 0.0:
                value = 0

    else:
        raise TypeError("Invalid type in KeywordSet ({type(value)}), only simple Python types are allowed.")

    return value


def hash_dictionary(data: Dict[str, Any]) -> str:
    m = hashlib.sha1()
    m.update(json.dumps(data, sort_keys=True).encode("UTF-8"))
    return m.hexdigest()



"""
Tests the server collection compute capabilities.
"""
import itertools
import pathlib
from contextlib import contextmanager
from typing import List

import numpy as np
import pandas as pd
import pytest
import qcelemental as qcel
from qcelemental.models import Molecule, ProtoModel

import qcfractal.interface as ptl
from qcengine.testing import is_program_new_enough
from qcfractal import testing
from qcfractal.testing import df_compare, fractal_compute_server, live_fractal_or_skip


@contextmanager
def check_requests_monitor(client, request, request_made=True, kind="get"):
    before = client._request_counter[(request, kind)]
    yield
    after = client._request_counter[(request, kind)]

    if request_made:
        assert after > before, f"Requests were were expected, but none were made. Request type: {request} ({kind})"
    else:
        assert after == before, f"Requests were made when none were expected. Request type: {request} ({kind})"


def handle_dataset_fixture_params(client, ds_type, ds, fractal_compute_server, request):
    ds = client.get_collection(ds_type, ds.name)
    if request.param == "no_view":
        ds._disable_view = True
    elif request.param == "download_view":
        ds._disable_view = False
        try:
            import requests_mock
        except ImportError:
            pytest.skip("Missing request_mock")
        with requests_mock.Mocker(real_http=True) as m:
            with open(fractal_compute_server.view_handler.view_path(ds.data.id), "rb") as f:
                m.get(ds.data.view_url_hdf5, body=f)
                ds.download(verify=True)
    elif request.param == "remote_view":
        ds._disable_view = False
        if ds._view is None:
            raise ValueError("Remote view is not available.")
        assert isinstance(ds._view, ptl.collections.RemoteView)
    else:
        raise ValueError(f"Unknown dataset fixture parameter: {request.param}.")
    return ds


def build_dataset_fixture_view(ds, fractal_compute_server):
    view = ptl.collections.HDF5View(fractal_compute_server.view_handler.view_path(ds.data.id))
    view.write(ds)
    ds.data.__dict__["view_available"] = True
    ds.data.__dict__["view_url_hdf5"] = f"http://mock.repo/{ds.data.id}/latest.hdf5"
    ds.data.__dict__["view_metadata"] = {"blake2b_checksum": view.hash()}
    ds.save()


@pytest.fixture(scope="module", params=["download_view", "no_view", "remote_view"])
def gradient_dataset_fixture(fractal_compute_server, tmp_path_factory, request):
    client = ptl.FractalClient(fractal_compute_server)

    try:
        ds = client.get_collection("Dataset", "ds_gradient")
    except KeyError:
        testing.check_has_module("psi4")

        # Build a dataset
        ds = ptl.collections.Dataset(
            "ds_gradient", client, default_program="psi4", default_driver="gradient", default_units="hartree/bohr"
        )

        ds.add_entry("He1", ptl.Molecule.from_data("He -1 0 0\n--\nHe 0 0 1"))
        ds.add_entry("He2", ptl.Molecule.from_data("He -1.1 0 0\n--\nHe 0 0 1.1"))
        ds.save()

        contrib = {
            "name": "Gradient",
            "theory_level": "pseudo-random values",
            "values": [[0.03, 0, 0.02, -0.02, 0, -0.03], [0.03, 0, 0.02, -0.02, 0, -0.03]],
            "index": ["He1", "He2"],
            "theory_level_details": {"driver": "gradient"},
            "units": "hartree/bohr",
        }
        contrib_no_details = {
            "name": "no details",
            "theory_level": "pseudo-random values",
            "values": [[0.03, 0, 0.02, -0.02, 0, -0.03], [0.03, 0, 0.02, -0.02, 0, -0.03]],
            "index": ["He1", "He2"],
            "units": "hartree/bohr",
        }
        contrib_all_details = {
            "name": "all details",
            "theory_level": "pseudo-random values",
            "values": [[0.03, 0, 0.02, -0.02, 0, -0.03], [0.03, 0, 0.02, -0.02, 0, -0.03]],
            "index": ["He1", "He2"],
            "theory_level_details": {
                "driver": "gradient",
                "program": "fake_program",
                "basis": "fake_basis",
                "method": "fake_method",
                "keywords": "fake_keywords",
            },
            "units": "hartree/bohr",
        }
        ds.add_contributed_values(contrib)
        ds.add_contributed_values(contrib_all_details)
        ds.add_contributed_values(contrib_no_details)

        ds.add_keywords("scf_default", "psi4", ptl.models.KeywordSet(values={}), default=True)
        ds.save()

        ds.compute("HF", "sto-3g")
        ds.compute("HF", "3-21g")
        fractal_compute_server.await_results()

        assert ds.get_records("HF", "sto-3g").iloc[0, 0].status == "COMPLETE"
        assert ds.get_records("HF", "sto-3g").iloc[1, 0].status == "COMPLETE"
        assert ds.get_records("HF", "3-21g").iloc[0, 0].status == "COMPLETE"
        assert ds.get_records("HF", "3-21g").iloc[1, 0].status == "COMPLETE"

        build_dataset_fixture_view(ds, fractal_compute_server)

    ds = handle_dataset_fixture_params(client, "Dataset", ds, fractal_compute_server, request)

    yield client, ds


def test_gradient_dataset_get_molecules(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture
    request_made = not ds._use_view(False)
    ds._clear_cache()

    he1_dist = 2.672476322216822
    he2_dist = 2.939723950195864
    with check_requests_monitor(client, "molecule", request_made=request_made):
        mols = ds.get_molecules()
    assert mols.shape == (2, 1)
    assert mols.iloc[0, 0].measure([0, 1]) == pytest.approx(he1_dist)
    assert mols.iloc[1, 0].measure([0, 1]) == pytest.approx(he2_dist)

    mol = ds.get_molecules(subset="He1")
    assert mol.measure([0, 1]) == pytest.approx(he1_dist)

    mol = ds.get_molecules(subset="He2")
    assert mol.measure([0, 1]) == pytest.approx(he2_dist)

    mols_subset = ds.get_molecules(subset=["He1"])
    assert mols_subset.iloc[0, 0].measure([0, 1]) == pytest.approx(he1_dist)

    with pytest.raises((KeyError, RuntimeError)):
        ds.get_molecules(subset="NotInDataset")


def test_gradient_dataset_get_records(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture

    with check_requests_monitor(client, "result", request_made=True):
        records = ds.get_records("HF", "sto-3g")

    assert records.shape == (2, 1)
    assert records.iloc[0, 0].status == "COMPLETE"
    assert records.iloc[1, 0].status == "COMPLETE"

    with check_requests_monitor(client, "result", request_made=True):
        records_subset1 = ds.get_records("HF", "sto-3g", subset="He2")
    assert records_subset1.status == "COMPLETE"

    with check_requests_monitor(client, "result", request_made=True):
        records_subset2 = ds.get_records("HF", "sto-3g", subset=["He2"])
    assert records_subset2.shape == (1, 1)
    assert records_subset2.iloc[0, 0].status == "COMPLETE"

    with check_requests_monitor(client, "result", request_made=True):
        rec_proj = ds.get_records("HF", "sto-3g", include=["extras", "return_result"])
    assert rec_proj.shape == (2, 2)
    assert set(rec_proj.columns) == {"extras", "return_result"}

    with pytest.raises(KeyError):
        with check_requests_monitor(client, "result", request_made=False):
            ds.get_records(method="NotInDataset")


def test_gradient_dataset_get_values(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture

    request_made = not ds._use_view(False)
    ds._clear_cache()

    with check_requests_monitor(client, "result", request_made=request_made):
        cols = set(ds.get_values().columns)
    names = set(ds.list_values().reset_index()["name"])
    assert cols == names

    df = ds.get_values()
    assert df.shape == (len(ds.get_index()), 5)
    for entry in ds.get_index():
        assert (
            df.loc[entry, "HF/sto-3g"]
            == ds.get_records(subset=entry, method="hf", basis="sto-3g", include=["return_result"])
        ).all()

    assert ds.get_values(method="NotInDataset").shape[1] == 0  # 0-length DFs can cause exceptions

    with pytest.warns(RuntimeWarning):
        ds.get_values(name="HF/sto-3g", basis="sto-3g")


def test_gradient_dataset_list_values(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture

    # List contributed values
    df = ds.list_values(native=False).reset_index()

    assert df.shape == (3, 7)
    assert set(df.columns) == {*ds.data.history_keys, "name", "native"}
    assert {x.lower() for x in df["name"]} == set(ds.data.contributed_values.keys())

    df = ds.list_values(driver="graDieNt", native=False).reset_index()
    assert {x.lower() for x in df["name"]} == {"gradient", "all details"}

    names = ["GrAdIeNt", "no details"]
    df = ds.list_values(name=names, native=False).reset_index()
    assert {x.lower() for x in df["name"]} == {name.lower() for name in names}

    # List native values
    df1 = ds.list_values(native=True).reset_index()
    assert df1.shape == (2, 7)
    assert set(df1.columns) == {*ds.data.history_keys, "name", "native"}

    df2 = ds.list_values(method="hf", basis="sto-3g", native=True).reset_index()
    df3 = ds.list_values(name="hf/sto-3g", native=True).reset_index()
    assert (df2 == df3).all().all()

    # All values
    df = ds.list_values().reset_index()
    assert df.shape == (5, 7)
    assert set(df.columns) == {*ds.data.history_keys, "name", "native"}

    df = ds.list_values(driver="gradient").reset_index()
    assert df.shape == (4, 7)

    df = ds.list_values(name="Not in dataset").reset_index()
    assert len(df) == 0


def test_gradient_dataset_statistics(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture

    df = ds.get_values(native=True)
    assert df.shape == (2, 2)
    assert np.sum(df.loc["He2", "HF/sto-3g"]) == pytest.approx(0.0)

    # Test out some statistics
    stats = ds.statistics("MUE", "HF/sto-3g", "Gradient")
    assert pytest.approx(stats.mean(), 1.0e-5) == 0.00984176986312362

    stats = ds.statistics("UE", "HF/sto-3g", "Gradient")
    assert pytest.approx(stats.loc["He1"].mean(), 1.0e-5) == 0.01635020639
    assert pytest.approx(stats.loc["He2"].mean(), 1.0e-5) == 0.00333333333


def test_gradient_dataset_get_values_caching(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture

    ds._clear_cache()

    with check_requests_monitor(client, "result", request_made=True and not ds._use_view(False)):
        ds.get_values()

    with check_requests_monitor(client, "result", request_made=False):
        ds.get_values()

    ds._clear_cache()

    with check_requests_monitor(client, "result", request_made=True and not ds._use_view(False)):
        ds.get_values(basis="sto-3g")

    with check_requests_monitor(client, "result", request_made=False):
        ds.get_values(basis="sto-3g", subset=["He1"])
        ds.get_values(basis="sto-3g", subset="He2")
        ds.get_values(basis="sto-3g", subset=["He1", "He2"])

    with check_requests_monitor(client, "result", request_made=True and not ds._use_view(False)):
        ds.get_values(basis="3-21g", subset="He1")

    with check_requests_monitor(client, "result", request_made=True and not ds._use_view(False)):
        ds.get_values(basis="3-21g", subset=["He1", "He2"])

    with check_requests_monitor(client, "result", request_made=False):
        ds.get_values(basis="3-21g", subset="He2")

    with check_requests_monitor(client, "result", request_made=False):
        ds.get_values()


@pytest.mark.parametrize("use_cache", [True, False])
def test_gradient_dataset_values_subset(gradient_dataset_fixture, use_cache):
    client, ds = gradient_dataset_fixture

    allvals = ds.get_values()
    ds._clear_cache()

    subsets = [None, "He1", ["He2"], ["He1", "He2"], ["He1", "He2", "He1", "He1"]]
    colnames = [None, "HF/3-21g", ["HF/3-21g", "HF/sto-3g"], "no details", ["HF/sto-3g", "no details"]]
    for subset, colname in itertools.product(subsets, colnames):
        if not use_cache:
            ds._clear_cache()
        df1 = ds.get_values(subset=subset, name=colname)

        if colname is None:
            c = slice(None)
        elif isinstance(colname, str):
            c = [colname]
        else:
            c = colname
        df2 = allvals.loc[subset if subset is not None else slice(None), c]
        assert df_compare(df1, df2, sort=True)


def test_gradient_dataset_records_args(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture

    assert len(ds.get_records(method="hf", keywords="scf_default", basis="sto-3g")) == 2
    assert (
        len(ds.get_records(method="hf", keywords="scf_default")) == 4
    )  # TODO: we might want a multi-keyword dataset fixture here
    assert len(ds.get_records(method="hf")) == 4


@pytest.fixture(scope="module", params=["download_view", "no_view", "remote_view"])
def contributed_dataset_fixture(fractal_compute_server, tmp_path_factory, request):
    """ Fixture for testing rich contributed datasets with many properties and molecules of different sizes"""
    client = ptl.FractalClient(fractal_compute_server)
    try:
        ds = client.get_collection("Dataset", "ds_contributed")
    except KeyError:

        # Build a dataset
        ds = ptl.collections.Dataset("ds_contributed", client)

        ds.add_entry("He1", ptl.Molecule.from_data("He -1 0 0\n--\nHe 0 0 1"))
        ds.add_entry("He", ptl.Molecule.from_data("He -1.1 0 0"))
        ds.units = "hartree"
        ds.save()

        energy = {
            "name": "Fake Energy",
            "theory_level": "pseudo-random values",
            "values": np.array([1234.5, 5.4321]),
            "index": ["He1", "He"],
            "theory_level_details": {
                "driver": "energy",
                "program": "fake_program",
                "basis": "fake_basis",
                "method": "fake_method",
            },
            "units": "hartree",
        }
        gradient = {
            "name": "Fake Gradient",
            "theory_level": "pseudo-random values",
            "values": [np.array([0.03, 0, 0.02, -0.02, 0, -0.03]), np.array([0.03, 0, 0.02])],
            "index": ["He1", "He"],
            "theory_level_details": {
                "driver": "gradient",
                "program": "fake_program",
                "basis": "fake_basis",
                "method": "fake_method",
            },
            "units": "hartree/bohr",
        }
        hessian = {
            "name": "Fake Hessian",
            "theory_level": "pseudo-random values",
            "values": [np.eye(6).ravel(), np.array([1, 0.2, 0.1, 0.2, 1, 0.4, 0.1, 0.4, 1])],
            "index": ["He1", "He"],
            "theory_level_details": {
                "driver": "hessian",
                "program": "fake_program",
                "basis": "fake_basis",
                "method": "fake_method",
            },
            "units": "hartree/bohr**2",
        }
        dipole = {
            "name": "Fake Dipole",
            "theory_level": "pseudo-random values",
            "values": [np.array([1.0, 2.0, 3.0]), np.array([1.0, -2.0, 0.0])],
            "index": ["He1", "He"],
            "theory_level_details": {
                "driver": "dipole",
                "program": "fake_program",
                "basis": "fake_basis",
                "method": "fake_method",
            },
            "units": "e * bohr",
        }
        energy_FF = {
            "name": "Fake FF Energy",
            "theory_level": "some force field",
            "values": [0.5, 42.0],
            "index": ["He1", "He"],
            "theory_level_details": {
                "driver": "energy",
                "program": "fake_program",
                "basis": None,
                "method": "fake_method",
            },
            "units": "hartree",
        }

        ds.add_contributed_values(energy)
        ds.add_contributed_values(gradient)
        ds.add_contributed_values(hessian)
        ds.add_contributed_values(dipole)
        ds.add_contributed_values(energy_FF)

        with pytest.raises(KeyError):
            ds.add_contributed_values(energy)

        ds.save()

        build_dataset_fixture_view(ds, fractal_compute_server)

    ds = handle_dataset_fixture_params(client, "Dataset", ds, fractal_compute_server, request)

    yield client, ds


def test_dataset_contributed_units(contributed_dataset_fixture):
    _, ds = contributed_dataset_fixture

    assert qcel.constants.ureg(
        ds._column_metadata[ds.get_values(name="Fake Energy").columns[0]]["units"]
    ) == qcel.constants.ureg(ds.units)
    assert qcel.constants.ureg(
        ds._column_metadata[ds.get_values(name="Fake Gradient").columns[0]]["units"]
    ) == qcel.constants.ureg("hartree/bohr")
    assert qcel.constants.ureg(
        ds._column_metadata[ds.get_values(name="Fake Hessian").columns[0]]["units"]
    ) == qcel.constants.ureg("hartree/bohr**2")
    assert qcel.constants.ureg(
        ds._column_metadata[ds.get_values(name="Fake Dipole").columns[0]]["units"]
    ) == qcel.constants.ureg("e * bohr")

    old_units = ds.units
    assert old_units == "kcal / mol"
    ds.units = "kcal/mol"
    before = ds.get_values()
    ds.units = "hartree"
    after = ds.get_values()
    assert before["Fake Energy"][0] == after["Fake Energy"][0] / qcel.constants.conversion_factor("kcal/mol", "hartree")
    assert before["Fake Gradient"][0][0, 0] == after["Fake Gradient"][0][0, 0]
    ds.units = old_units


def test_dataset_contributed_mixed_values(contributed_dataset_fixture):
    _, ds = contributed_dataset_fixture

    unselected_values = ds.get_values()
    assert unselected_values.shape == (2, 5)
    selected_values = ds.get_values(program="fake_program")
    assert selected_values.shape == (2, 5)
    selected_values = ds.get_values(basis="None")
    assert selected_values.shape == (2, 1)
    assert selected_values.columns[0] == "Fake FF Energy"


def test_dataset_compute_response(fractal_compute_server):
    """ Tests that the full compute response is returned when calling Dataset.compute """
    client = ptl.FractalClient(fractal_compute_server)

    # Build a dataset
    ds = ptl.collections.Dataset("ds", client, default_program="psi4", default_driver="energy", default_units="hartree")

    ds.add_entry("He1", ptl.Molecule.from_data("He -1 0 0\n--\nHe 0 0 1"))
    ds.add_entry("He2", ptl.Molecule.from_data("He -1.1 0 0\n--\nHe 0 0 1.1"))

    ds.save()

    # Compute fewer molecules than query limit
    response = ds.compute("HF", "sto-3g")
    assert len(response.ids) == 2

    # Compute more molecules than query limit
    client.query_limit = 1
    response = ds.compute("HF", "sto-3g")
    assert len(response.ids) == 2


@testing.using_psi4
def test_dataset_protocols(fractal_compute_server):
    """ Tests using protocols with dataset compute."""
    client = ptl.FractalClient(fractal_compute_server)

    # Build basis dataset
    ds = ptl.collections.Dataset("protocol_dataset", client, default_program="psi4", default_driver="energy")

    ds.add_entry("He1", ptl.Molecule.from_data("He 0 0 0\n--\nHe 0 0 2.2"))
    ds.save()

    # compute the wavefunction
    response = ds.compute(method="hf", basis="sto-3g", protocols={"wavefunction": "orbitals_and_eigenvalues"})

    # await the result and check for orbitals
    fractal_compute_server.await_results()

    result = client.query_results(id=response.ids)[0]
    orbitals = result.get_wavefunction("orbitals_a")
    assert orbitals.shape == (2, 2)

    basis = result.get_wavefunction("basis")
    assert basis.name.lower() == "sto-3g"


def test_reactiondataset_check_state(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)
    ds = ptl.collections.ReactionDataset("check_state", client, ds_type="ie", default_program="rdkit")
    ds.add_ie_rxn("He1", ptl.Molecule.from_data("He -3 0 0\n--\nNe 0 0 2"))

    with pytest.raises(ValueError):
        ds.compute("SCF", "STO-3G")

    with pytest.raises(ValueError):
        ds.get_records("SCF", "STO-3G")

    ds.save()

    ds.add_keywords("default", "psi4", ptl.models.KeywordSet(values={"a": 5}))

    with pytest.raises(ValueError):
        ds.get_records("SCF", "STO-3G")

    ds.save()

    contrib = {
        "name": "Benchmark",
        "doi": None,
        "theory_level": "very high",
        "values": [-0.00001098794749, 0.0009608501557],
        "index": ["He2", "He1"],
        "units": "hartree",
    }
    # No entry for He2 in the dataset
    with pytest.raises(ValueError):
        ds.add_contributed_values(contrib)

    ds.add_ie_rxn("He2", ptl.Molecule.from_data("He -2 0 0\n--\nNe 0 0 2"))
    ds.save()

    ds.add_contributed_values(contrib)

    with pytest.raises(ValueError):
        ds.get_records("SCF", "STO-3G")

    assert "benchmark" == ds.list_values(native=False).reset_index()["name"][0].lower()
    ds.units = "hartree"
    bench = ds.get_values(name="benchmark", native=False)
    assert ds._column_metadata[bench.columns[0]]["native"] is False
    assert bench.shape == (2, 1)
    assert bench.loc["He1"][0] == contrib["values"][1]
    assert bench.loc["He2"][0] == contrib["values"][0]


@pytest.mark.parametrize("use_cache", [True, False])
def test_contributed_dataset_values_subset(contributed_dataset_fixture, use_cache):
    client, ds = contributed_dataset_fixture

    ds._clear_cache()
    allvals = ds.get_values()
    ds._clear_cache()

    subsets = [None, "He1", ["He"], ["He", "He1"]]
    colnames = [None, "Fake Energy", ["Fake Hessian", "Fake Dipole", "Fake Energy"]]
    for subset, colname in itertools.product(subsets, colnames):
        if not use_cache:
            ds._clear_cache()
        df1 = ds.get_values(subset=subset, name=colname)

        if colname is None:
            c = slice(None)
        elif isinstance(colname, str):
            c = [colname]
        else:
            c = colname
        df2 = allvals.loc[subset if subset is not None else slice(None), c]
        assert df_compare(df1, df2, sort=True)


@pytest.fixture(scope="module", params=["no_view", "remote_view", "download_view"])
def reactiondataset_dftd3_fixture_fixture(fractal_compute_server, tmp_path_factory, request):
    ds_name = "He_DFTD3"
    client = ptl.FractalClient(fractal_compute_server)

    try:
        ds = client.get_collection("ReactionDataset", ds_name)
    except KeyError:
        testing.check_has_module("psi4")
        testing.check_has_module("dftd3")

        ds = ptl.collections.ReactionDataset(ds_name, client, ds_type="ie", default_units="hartree")

        # Add two helium dimers to the DB at 4 and 8 bohr
        HeDimer = ptl.Molecule.from_data([[2, 0, 0, -1.412], [2, 0, 0, 1.412]], dtype="numpy", units="bohr", frags=[1])
        ds.add_ie_rxn("HeDimer", HeDimer, attributes={"r": 4})
        ds.set_default_program("psi4")
        ds.add_keywords("scf_default", "psi4", ptl.models.KeywordSet(values={"e_convergence": 1.0e-10}), default=True)

        ds.save()

        ncomp1 = ds.compute("B3LYP-D3", "6-31g")
        assert len(ncomp1.ids) == 4
        assert len(ncomp1.submitted) == 4  # Dimer/monomer, dft/dftd3

        ncomp1 = ds.compute("B3LYP-D3", "6-31g", stoich="cp")
        assert len(ncomp1.ids) == 4
        assert len(ncomp1.submitted) == 2  # monomer, dft/dftd3

        ncomp2 = ds.compute("B3LYP-D3(BJ)", "6-31g")
        assert len(ncomp2.ids) == 4
        assert len(ncomp2.submitted) == 2  # dimer/monomer, dftd3

        ncomp1 = ds.compute("B3LYP-D3(BJ)", "6-31g", stoich="cp")
        assert len(ncomp1.ids) == 4
        assert len(ncomp1.submitted) == 1  # monomer, dftd3

        fractal_compute_server.await_results()

        build_dataset_fixture_view(ds, fractal_compute_server)

    ds = handle_dataset_fixture_params(client, "ReactionDataset", ds, fractal_compute_server, request)

    yield client, ds


def test_reactiondataset_dftd3_records(reactiondataset_dftd3_fixture_fixture):
    client, ds = reactiondataset_dftd3_fixture_fixture

    records = ds.get_records("B3LYP", "6-31g")
    assert records.shape == (1, 1)
    assert records.iloc[0, 0].status == "COMPLETE"

    records = ds.get_records("B3LYP", "6-31g", stoich="default")
    assert records.shape == (1, 1)
    assert records.iloc[0, 0].status == "COMPLETE"

    records = ds.get_records("B3LYP", "6-31g", stoich="default", subset="HeDimer")
    assert records.shape == (1, 1)

    # No molecules
    with pytest.raises(KeyError):
        records = ds.get_records("B3LYP", "6-31g", stoich="default", subset="Gibberish")

    # Bad stoichiometry
    with pytest.raises(KeyError):
        ds.get_records("B3LYP", "6-31g", stoich="cp5")


def test_reactiondataset_dftd3_energies(reactiondataset_dftd3_fixture_fixture):
    client, ds = reactiondataset_dftd3_fixture_fixture

    request_made = not ds._use_view(False)
    ds._clear_cache()

    bench = {
        "B3LYP/6-31g": pytest.approx(0.01822563, abs=1.0e-5),
        "cp-B3LYP/6-31g": pytest.approx(0.01867427, abs=1.0e-5),
        "B3LYP-D3/6-31g": pytest.approx(0.01815022, abs=1.0e-5),
        "cp-B3LYP-D3/6-31g": pytest.approx(0.01859886, abs=1.0e-5),
        "B3LYP-D3(BJ)/6-31g": pytest.approx(0.01814335, abs=1.0e-5),
        "cp-B3LYP-D3(BJ)/6-31g": pytest.approx(0.01859199, abs=1.0e-5),
    }

    with check_requests_monitor(client, "result", request_made=request_made):
        ret = ds.get_values("B3LYP", "6-31G")
    assert ret.loc["HeDimer", "B3LYP/6-31g"] == bench["B3LYP/6-31g"]

    ret = ds.get_values("B3LYP", "6-31G", stoich="cp")
    assert ret.loc["HeDimer", "cp-B3LYP/6-31g"] == bench["cp-B3LYP/6-31g"]

    ret = ds.get_values("B3LYP-D3", "6-31G")
    assert ret.loc["HeDimer", "B3LYP-D3/6-31g"] == bench["B3LYP-D3/6-31g"]

    ret = ds.get_values("B3LYP-D3(BJ)", "6-31G")
    assert ret.loc["HeDimer", "B3LYP-D3(BJ)/6-31g"] == bench["B3LYP-D3(BJ)/6-31g"]

    # Should be in ds.df now as well
    ds.get_values(stoich="cp")
    for key, value in bench.items():
        assert value == ds.df.loc["HeDimer", key]


def test_reactiondataset_dftd3_molecules(reactiondataset_dftd3_fixture_fixture):
    client, ds = reactiondataset_dftd3_fixture_fixture

    request_made = not ds._use_view(False)
    ds._clear_cache()

    with check_requests_monitor(client, "molecule", request_made=request_made):
        mols = ds.get_molecules()
    assert mols.shape == (1, 1)
    assert np.all(mols.iloc[0, 0].real)  # Should be all real
    assert tuple(mols.index) == (("HeDimer", "default", 0),)

    mols = ds.get_molecules(stoich="cp1")
    assert mols.shape == (1, 1)
    assert not np.all(mols.iloc[0, 0].real)  # Should be half real
    assert tuple(mols.index) == (("HeDimer", "cp1", 0),)

    stoichs = ["cp1", "default1", "cp", "default"]
    mols = ds.get_molecules(stoich=stoichs)
    assert mols.shape == (4, 1)

    mols = mols.reset_index()
    assert set(stoichs) == set(mols["stoichiometry"])


def test_rectiondataset_dftd3_values_caching(reactiondataset_dftd3_fixture_fixture):
    client, ds = reactiondataset_dftd3_fixture_fixture
    ds._clear_cache()

    with check_requests_monitor(client, "result", request_made=True and not ds._use_view(False)):
        ds.get_values("B3LYP", "6-31G")

    with check_requests_monitor(client, "result", request_made=True and not ds._use_view(False)):
        ds.get_values("B3LYP-D3", "6-31G")

    with check_requests_monitor(client, "result", request_made=True and not ds._use_view(False)):
        ds.get_values("B3LYP-D3(BJ)", "6-31G")

    with check_requests_monitor(client, "result", request_made=False):
        ds.get_values("B3LYP", "6-31G", subset=None)
        ds.get_values("B3LYP", "6-31G", subset="HeDimer")
        ds.get_values("B3LYP", "6-31G", subset=["HeDimer"])


@pytest.mark.parametrize("use_cache", [True, False])
def test_reactiondataset_dftd3_values_subset(reactiondataset_dftd3_fixture_fixture, use_cache):
    client, ds = reactiondataset_dftd3_fixture_fixture

    allvals = ds.get_values()
    ds._clear_cache()

    subsets = [None, "HeDimer", ["HeDimer"], ["HeDimer", "HeDimer"]]
    colnames = [None, "B3LYP/6-31g", ["B3LYP-D3/6-31g", "B3LYP-D3(BJ)/6-31g"], ["B3LYP/6-31g"]]
    for subset, colname in itertools.product(subsets, colnames):
        if not use_cache:
            ds._clear_cache()
        df1 = ds.get_values(subset=subset, name=colname)

        if colname is None:
            c = slice(None)
        elif isinstance(colname, str):
            c = [colname]
        else:
            c = colname
        df2 = allvals.loc[subset if subset is not None else slice(None), c]
        assert df_compare(df1, df2, sort=True)


def test_dataset_dftd3(reactiondataset_dftd3_fixture_fixture):
    client, rxn_ds = reactiondataset_dftd3_fixture_fixture

    if not rxn_ds._use_view():
        ds_name = "He_DFTD3"
        ds = ptl.collections.Dataset(ds_name, client)

        HeDimer = rxn_ds.get_molecules(subset="HeDimer").iloc[0, 0]
        ds.add_entry("HeDimer", HeDimer)
        ds.set_default_program(rxn_ds.data.default_program)
        ds.add_keywords("scf_default", rxn_ds.data.default_program, ptl.models.KeywordSet(values={}), default=True)

        ds.save()

        ds.compute("B3LYP-D3", "6-31g")

        d3 = ds.get_values(method="b3lyp-d3")
        assert d3.shape == (1, 1)
        b3lyp = ds.get_values(method="b3lyp")
        assert b3lyp.shape == (1, 1)

        assert d3.iloc[0, 0] != b3lyp.iloc[0, 0]


@testing.using_psi4
def test_compute_reactiondataset_regression(fractal_compute_server):
    """
    Tests an entire server and interaction energy dataset run
    """

    client = ptl.FractalClient(fractal_compute_server)
    ds_name = "He_PES"
    ds = ptl.collections.ReactionDataset(ds_name, client, ds_type="ie")

    # Add two helium dimers to the DB at 4 and 8 bohr
    He1 = ptl.Molecule.from_data([[2, 0, 0, -2], [2, 0, 0, 2]], dtype="numpy", units="bohr", frags=[1])
    ds.add_ie_rxn("He1", He1, attributes={"r": 4})

    # Save the DB and re-acquire via classmethod
    r = ds.save()
    ds = ptl.collections.ReactionDataset.from_server(client, ds_name)
    ds.set_default_program("psi4")
    assert "ReactionDataset(" in str(ds)

    # Test collection lists
    ret = client.list_collections(aslist=True)
    assert ds_name in ret["ReactionDataset"]

    ret = client.list_collections("reactiondataset", aslist=True)
    assert ds_name in ret

    He2 = ptl.Molecule.from_data([[2, 0, 0, -4], [2, 0, 0, 4]], dtype="numpy", units="bohr", frags=[1])
    ds.add_ie_rxn("He2", He2, attributes={"r": 4})
    ds.save()

    contrib = {
        "name": "Benchmark",
        "doi": None,
        "theory_level": "very high",
        "values": [0.0009608501557, -0.00001098794749],
        "index": ["He1", "He2"],
        "units": "hartree",
    }
    ds.add_contributed_values(contrib)
    ds.set_default_benchmark("Benchmark")

    # Save the DB and overwrite the result, reacquire via client
    ds.save()
    ds = client.get_collection("reactiondataset", ds_name)

    with pytest.raises(KeyError):
        ds.compute("SCF", "STO-3G", stoich="nocp")  # Should be 'default' not 'nocp'

    # Compute SCF/sto-3g
    ret = ds.compute("SCF", "STO-3G")
    assert len(ret.submitted) == 3
    fractal_compute_server.await_results()

    # Query computed results
    ret = ds.get_values("SCF", "STO-3G")
    assert ret.shape == (2, 1)
    assert pytest.approx(0.6024530476, 1.0e-5) == ret.loc["He1", "SCF/sto-3g"]
    assert pytest.approx(-0.0068950359, 1.0e-5) == ret.loc["He2", "SCF/sto-3g"]

    # Check results
    assert pytest.approx(0.00024477933196125805, 1.0e-5) == ds.statistics("MUE", "SCF/sto-3g")

    assert pytest.approx([0.081193, 7.9533e-05], 1.0e-1) == list(ds.statistics("URE", "SCF/sto-3g"))
    assert pytest.approx(0.0406367, 1.0e-5) == ds.statistics("MURE", "SCF/sto-3g")
    assert pytest.approx(0.002447793, 1.0e-5) == ds.statistics("MURE", "SCF/sto-3g", floor=10)

    assert isinstance(ds.to_json(), dict)
    assert ds.list_records(keywords=None).shape[0] == 1

    ds.units = "eV"
    assert pytest.approx(0.00010614635, 1.0e-5) == ds.statistics("MURE", "SCF/sto-3g", floor=10)

    # Check get_molecules
    mols = ds.get_molecules()
    assert mols.shape == (2, 1)

    mols = ds.get_molecules(stoich="cp1")
    assert mols.shape == (2, 1)


@testing.using_psi4
def test_compute_reactiondataset_keywords(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    mol1 = ptl.Molecule.from_data("He 0 0 -1.1\n--\nHe 0 0 1.1")

    # Build a dataset
    ds = ptl.collections.ReactionDataset("dataset_options", client, ds_type="ie")
    ds.set_default_program("Psi4")

    ds.add_ie_rxn("He2", mol1)
    ds.add_keywords("direct", "psi4", ptl.models.KeywordSet(values={"scf_type": "direct"}), default=True)
    ds.add_keywords("df", "psi4", ptl.models.KeywordSet(values={"scf_type": "df"}))

    ds.save()
    ds = client.get_collection("reactiondataset", "dataset_options")

    # Compute, should default to direct options
    r = ds.compute("SCF", "STO-3G")
    fractal_compute_server.await_results()
    ret = ds.get_values("SCF", "STO-3G")
    assert pytest.approx(0.39323818102293856, 1.0e-5) == ds.df.loc["He2", "SCF/sto-3g"]

    r = ds.compute("SCF", "sto-3g", keywords="df")
    fractal_compute_server.await_results()
    ds.get_values("SCF", "sto-3g", keywords="df").columns[0] == "SCF/sto-3g-df"
    assert pytest.approx(0.38748602675524185, 1.0e-5) == ds.df.loc["He2", "SCF/sto-3g-df"]

    assert ds.list_records().shape[0] == 2
    assert ds.list_records(keywords="DF").shape[0] == 1
    assert ds.list_records(keywords="DIRECT").shape[0] == 1

    # Check saved history
    ds = client.get_collection("reactiondataset", "dataset_options")
    assert ds.list_records().shape[0] == 2
    assert {"df", "direct"} == set(ds.list_records().reset_index()["keywords"])

    # Check keywords
    kw = ds.get_keywords("df", "psi4")
    assert kw.values["scf_type"] == "df"


@pytest.fixture(scope="module", params=[True, False], ids=["local_view", "no_view"])
def qm3_fixture(request, tmp_path_factory):
    # Connect to the QCArchive
    client = live_fractal_or_skip()
    ds = client.get_collection("Dataset", "QM3")
    ds._disable_query_limit = True

    # Trim down dataset for faster test
    to_remove = {row for row in ds.data.history if row[2].lower() not in {"b3lyp", "pbe"}}
    for row in to_remove:
        ds.data.history.remove(row)
    ds.get_entries(force=True)

    # with view
    if request.param:
        view = ptl.collections.HDF5View(pathlib.Path(tmp_path_factory.mktemp("test_collections"), "ds_qm3.hdf5"))
        view.write(ds)
        ds._view = view
    else:
        ds._view = None

    yield client, ds


@pytest.fixture(scope="module", params=[True, False], ids=["local_view", "no_view"])
def s22_fixture(request, tmp_path_factory):
    # Connect to the QCArchive
    client = live_fractal_or_skip()
    ds = client.get_collection("ReactionDataset", "S22")
    ds._disable_query_limit = True

    # Trim down dataset for faster test
    to_remove = {row for row in ds.data.history if row[2].lower() not in {"b3lyp", "pbe"}}
    for row in to_remove:
        ds.data.history.remove(row)
    ds.get_entries(force=True)

    # with view
    if request.param:
        view = ptl.collections.HDF5View(pathlib.Path(tmp_path_factory.mktemp("test_collections"), "ds_s22.hdf5"))
        view.write(ds)
        ds._view = view
    else:
        ds._view = None

    yield client, ds


@pytest.mark.slow
def test_qm3_list_select(qm3_fixture):
    """ tests list_values and get_values with multiple selections on the method and basis field """
    client, ds = qm3_fixture

    methods = {"b3lyp", "pbe"}
    bases = {"def2-svp", "def2-tzvp"}
    names = {f"{method}/{basis}" for (method, basis) in itertools.product(methods, bases)}

    df = ds.list_values(method=["b3lyp", "pbe"], basis=["def2-svp", "def2-tzvp"]).reset_index()
    assert names == set(df["name"].str.lower())

    df = ds.get_values(method=["b3lyp", "pbe"], basis=["def2-svp", "def2-tzvp"])
    assert names == set(df.columns.str.lower())

    df = ds.list_values(name=list(names)).reset_index()
    assert names == set(df["name"].str.lower())

    df = ds.get_values(name=list(names))
    assert names == set(df.columns.str.lower())


@pytest.mark.slow
def test_s22_list_select(s22_fixture):
    """ tests list_values and get_values with multiple selections on the method and basis field """
    client, ds = s22_fixture

    methods = {"b3lyp", "pbe"}
    bases = {"def2-svp", "def2-tzvp"}
    names = {f"{method}/{basis}" for (method, basis) in itertools.product(methods, bases)}

    df = ds.list_values(method=["b3lyp", "pbe"], basis=["def2-svp", "def2-tzvp"]).reset_index()
    assert names == set(df[df["stoichiometry"] == "default"]["name"].str.lower())

    df = ds.get_values(method=["b3lyp", "pbe"], basis=["def2-svp", "def2-tzvp"], stoich="default")
    assert names == set(df.columns.str.lower())

    df = ds.get_values(name=list(names))
    assert names == set(df.columns.str.lower())


@testing.using_psi4
@testing.using_dftd3
def test_rds_rxn(fractal_compute_server):
    client = fractal_compute_server.client()
    ds = ptl.collections.ReactionDataset("rds_rxn", client, "rxn")
    HeDimer1 = ptl.Molecule.from_data([[2, 0, 0, -1.412], [2, 0, 0, 1.412]], dtype="numpy", units="bohr", frags=[1])
    HeDimer2 = ptl.Molecule.from_data([[2, 3, 3, -1.412], [2, 3, 3, 1.412]], dtype="numpy", units="bohr", frags=[1])

    ds.add_rxn("HeDimer1", stoichiometry={"default": [(HeDimer1, 0.0)]})
    ds.add_rxn("HeDimer2", stoichiometry={"default": [(HeDimer2, 1.0), (HeDimer1, -1.0)]})
    ds.set_default_program("psi4")
    ds.add_keywords("scf_default", "psi4", ptl.models.KeywordSet(values={"e_convergence": 1.0e-10}), default=True)

    ds.save()

    ds.compute("B3LYP-D3", "6-31g")
    ds.get_values(method="B3LYP-D3", basis="6-31g")


def assert_list_get_values(ds):
    """ Tests that the output of list_values can be used as input to get_values"""
    columns = ds.list_values().reset_index()
    all_specs_unique = len(columns.drop("name", axis=1).drop_duplicates()) == len(columns.drop("name", axis=1))
    for row in columns.to_dict("records"):
        spec = row.copy()
        name = spec.pop("name")
        if "stoichiometry" in spec:
            spec["stoich"] = spec.pop("stoichiometry")
        from_name = ds.get_values(name=name)
        from_spec = ds.get_values(**spec)
        assert from_name.shape == (len(ds.get_index()), 1)
        if not all_specs_unique:
            continue
        assert from_spec.shape == (len(ds.get_index()), 1)
        assert from_name.columns[0] == from_spec.columns[0]


def test_gradient_dataset_list_get_values(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture
    assert_list_get_values(ds)


def test_contributed_dataset_list_get_values(contributed_dataset_fixture):
    client, ds = contributed_dataset_fixture
    assert_list_get_values(ds)


def test_d3_dataset_list_get_values(reactiondataset_dftd3_fixture_fixture):
    client, ds = reactiondataset_dftd3_fixture_fixture
    assert_list_get_values(ds)


@pytest.mark.slow
def test_qm3_list_get_values(qm3_fixture):
    client, ds = qm3_fixture
    assert_list_get_values(ds)


@pytest.mark.slow
def test_s22_list_get_values(s22_fixture):
    client, ds = s22_fixture
    assert_list_get_values(ds)


def assert_view_identical(ds):
    """ Tests if get_values, list_values, get_entries, and get_molecules return the same result with/out a view"""

    ds._disable_view = True
    list_ds = ds.list_values(force=True)
    cv_ds = ds.get_values(native=False, force=True)
    nv_ds = ds.get_values(native=True, force=True)
    v_ds = ds.get_values(force=True)
    entry_ds = ds.get_entries(force=True)
    mol_ds = ds.get_molecules(force=True)

    ds._disable_view = False
    list_view = ds.list_values()
    cv_view = ds.get_values(native=False)
    nv_view = ds.get_values(native=True)
    v_view = ds.get_values()
    entry_view = ds.get_entries()
    mol_view = ds.get_molecules()

    # 'no details' column in gradient_dataset_fixture does not match
    # because there is not enough info to interpret it as a gradient
    if "no details" in cv_view.columns:
        cv_view.drop("no details", axis=1, inplace=True)
        cv_ds.drop("no details", axis=1, inplace=True)
        v_view.drop("no details", axis=1, inplace=True)
        v_ds.drop("no details", axis=1, inplace=True)

    # Molecule IDs are different on the server and in the view by design
    for mid in ["molecule", "molecule_id"]:
        if mid in entry_ds.columns:
            entry_ds.drop(mid, axis=1, inplace=True)
            entry_view.drop(mid, axis=1, inplace=True)

    assert list_ds.equals(list_view)
    assert df_compare(cv_view, cv_ds)
    assert df_compare(nv_view, nv_ds)
    assert df_compare(v_view, v_ds)
    assert df_compare(entry_ds, entry_view)
    assert df_compare(mol_ds, mol_view)


def test_gradient_dataset_view_identical(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture
    assert_view_identical(ds)


def test_contributed_dataset_view_identical(contributed_dataset_fixture):
    client, ds = contributed_dataset_fixture
    assert_view_identical(ds)


def test_d3_dataset_view_identical(reactiondataset_dftd3_fixture_fixture):
    client, ds = reactiondataset_dftd3_fixture_fixture
    assert_view_identical(ds)


@pytest.mark.slow
def test_qm3_view_identical(qm3_fixture):
    client, ds = qm3_fixture
    assert_view_identical(ds)


@pytest.mark.slow
def test_s22_view_identical(s22_fixture):
    client, ds = s22_fixture
    assert_view_identical(ds)


@pytest.mark.slow
def test_view_download_remote(s22_fixture):
    _, ds = s22_fixture

    ds.data.__dict__["view_url_hdf5"] = "https://github.com/mattwelborn/QCArchiveViews/raw/master/S22/latest.hdf5"
    ds.data.__dict__["view_metadata"] = {
        "blake2b_checksum": "f9d537a982f63af0c500753b0e4779604252b9289fa26b6d83b657bf6e1039f1af2e44bbc819001fb857f1602280892b48422b8649cea786cdec3d2eb73412a9"
    }
    ds.download()  # 800 kb

    _, dsgz = s22_fixture

    dsgz.data.__dict__["view_url_hdf5"] = "https://github.com/mattwelborn/QCArchiveViews/raw/master/S22/latest.hdf5.gz"
    dsgz.data.__dict__["view_metadata"] = {
        "blake2b_checksum": "f9d537a982f63af0c500753b0e4779604252b9289fa26b6d83b657bf6e1039f1af2e44bbc819001fb857f1602280892b48422b8649cea786cdec3d2eb73412a9"
    }
    dsgz.download()  # 90 kb

    assert ds._view.hash() == dsgz._view.hash()


def test_view_download_mock(gradient_dataset_fixture, tmp_path_factory):
    try:
        import requests_mock
    except ImportError:
        pytest.skip("Missing request_mock")

    client, ds = gradient_dataset_fixture

    with requests_mock.Mocker(real_http=True) as m:
        path = pathlib.Path(tmp_path_factory.mktemp("test_collections"), "ds_gradient_remote.hdf5")
        ds.to_file(path, "hdf5")

        fake_url = "https://qcarchiveviews.com/gradient_ds.h5"
        ds.data.__dict__["view_url_hdf5"] = fake_url
        assert ds.data.id == ds.save()

        with open(path, "rb") as f:
            m.get(fake_url, body=f)
            ds = client.get_collection("Dataset", ds.name)
            ds.download(verify=False)

            # Check main functions run
            ds.get_entries()
            ds.list_values()

            with check_requests_monitor(client, "molecule", request_made=False):
                ds.get_molecules()

            with check_requests_monitor(client, "record", request_made=False):
                ds.get_values()

            ds.data.__dict__["view_metadata"] = {"blake2b_checksum": "badhash"}
            with pytest.raises(ValueError):
                ds.download(verify=True)


def test_gradient_dataset_plaintextview_write(gradient_dataset_fixture, tmpdir):
    _, ds = gradient_dataset_fixture
    ds.to_file(tmpdir / "test.tar.gz", "plaintext")


def test_contributed_dataset_plaintextview_write(contributed_dataset_fixture, tmpdir):
    _, ds = contributed_dataset_fixture
    ds.to_file(tmpdir / "test.tar.gz", "plaintext")


@pytest.mark.slow
def test_s22_dataset_plaintextview_write(s22_fixture, tmpdir):
    _, ds = s22_fixture
    ds.to_file(tmpdir / "test.tar.gz", "plaintext")


@pytest.mark.slow
def test_qm3_dataset_plaintextview_write(qm3_fixture, tmpdir):
    _, ds = qm3_fixture
    ds.to_file(tmpdir / "test.tar.gz", "plaintext")


def test_reactiondataset_dftd3_dataset_plaintextview_write(reactiondataset_dftd3_fixture_fixture, tmpdir):
    _, ds = reactiondataset_dftd3_fixture_fixture
    ds.to_file(tmpdir / "test.tar.gz", "plaintext")


@pytest.mark.slow
def test_s22_dataset_get_molecules_subset(s22_fixture):
    _, ds = s22_fixture
    ds.get_molecules(subset="Adenine-Thymine Complex WC")


### Non-dataset tests


def test_collection_query(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)

    ds = ptl.collections.Dataset("CAPITAL", client)
    ds.save()

    cols = client.list_collections()
    assert ("Dataset", "CAPITAL") in cols.index

    ds = client.get_collection("dataset", "capital")
    assert ds.name == "CAPITAL"

    ds = client.get_collection("DATAset", "CAPital")
    assert ds.name == "CAPITAL"


def test_generic_collection(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)
    g = ptl.collections.Generic("Generic1", client=client)

    # Check getters/savers
    g["hello"] = 5
    assert g["hello"] == 5

    # Copy the data over
    cp = g.get_data()
    assert cp.data["hello"] == 5
    cp.data["hello"] = 10

    # Make sure we did not change underlying data.
    assert g["hello"] == 5

    # Save semantics slightly wonky, double check this works
    g.save()
    g2 = ptl.collections.Generic.from_server(client, "Generic1")
    assert g2["hello"] == 5


def test_missing_collection(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)
    with pytest.raises(KeyError):
        client.get_collection("reactiondataset", "_waffles_")


@pytest.mark.slow
@testing.using_torsiondrive
@testing.using_geometric
@testing.using_rdkit
def test_torsiondrive_dataset(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    ds = ptl.collections.TorsionDriveDataset("testing", client=client)

    hooh1 = ptl.data.get_molecule("hooh.json")
    hooh2 = hooh1.copy(update={"geometry": hooh1.geometry + np.array([0, 0, 0.2])})

    ds.add_entry("hooh1", [hooh1], [[0, 1, 2, 3]], [60], attributes={"something": "hooh1"})
    ds.add_entry("hooh2", [hooh2], [[0, 1, 2, 3]], [60], attributes={"something": "hooh2"})

    optimization_spec = {"program": "geometric", "keywords": {"coordsys": "tric"}}
    qc_spec = {"driver": "gradient", "method": "UFF", "basis": None, "keywords": None, "program": "rdkit"}

    ds.add_specification("Spec1", optimization_spec, qc_spec, description="This is a really cool spec")

    ncompute = ds.compute("spec1")
    assert ncompute == 2
    assert ds.status("Spec1")["Spec1"].sum() == 2

    ds.save()

    fractal_compute_server.await_services(max_iter=1)

    # Check status
    ds.query("Spec1", force=True)
    status_basic = ds.status()
    assert status_basic.loc["RUNNING", "Spec1"] == 2
    status_spec = ds.status("Spec1")
    assert status_basic.loc["RUNNING", "Spec1"] == 2

    status_detail = ds.status("Spec1", detail=True)
    assert status_detail.loc["hooh2", "Complete Tasks"] == 1
    assert status_detail.loc["hooh2", "Total Points"] == 6

    # List of length 1 with detail=True
    status_detail = ds.status(["Spec1"], detail=True)
    assert status_detail.loc["hooh2", "Complete Tasks"] == 1
    assert status_detail.loc["hooh2", "Total Points"] == 6

    fractal_compute_server.await_services(max_iter=7)

    ds = client.get_collection("torsiondrivedataset", "testing")
    ds.query("spec1")

    # Add another fake set, should instantly return
    ds.add_specification("spec2", optimization_spec, qc_spec, description="This is a really cool spec")

    # Test subsets
    ncompute = ds.compute("spec2", subset=set())
    assert ncompute == 0

    ncompute = ds.compute("spec2")
    assert ncompute == 2
    ds.query("spec2")

    # We effectively computed the same thing twice with two duplicate specs
    for row in ["hooh1", "hooh2"]:
        for spec in ["Spec1", "spec2"]:
            assert pytest.approx(ds.df.loc[row, spec].get_final_energies(60), 1.0e-5) == 0.0007991272305133664

    assert ds.status().loc["COMPLETE", "Spec1"] == 2
    assert ds.status(collapse=False).loc["hooh1", "Spec1"] == "COMPLETE"

    assert ds.counts("hooh1").loc["hooh1", "Spec1"] > 5
    assert ds.counts("hooh1", specs="spec1", count_gradients=True).loc["hooh1", "Spec1"] > 30


def test_dataset_list_keywords(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)

    kw1 = ptl.models.KeywordSet(values={"foo": True})
    kw2 = ptl.models.KeywordSet(values={"foo": False})
    kw3 = ptl.models.KeywordSet(values={"foo": 13})
    ds = ptl.collections.Dataset("test_dataset_list_keywords", client)

    ds.add_keywords(alias="p1_k1", program="p1", keyword=kw1)
    ds.add_keywords(alias="p1_k2", program="p1", keyword=kw2, default=True)
    ds.add_keywords(alias="p2_k3", program="p2", keyword=kw3)
    ds.add_keywords(alias="p3_k1", program="p3", keyword=kw1, default=True)
    ds.save()

    res = ds.list_keywords().reset_index().drop("id", axis=1)
    ref = pd.DataFrame(
        {
            "program": ["p1", "p1", "p2", "p3"],
            "keywords": ["p1_k1", "p1_k2", "p2_k3", "p3_k1"],
            "default": [False, True, False, True],
        }
    )
    assert df_compare(res, ref), res


@testing.using_geometric
@testing.using_rdkit
def test_optimization_dataset(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    ds = ptl.collections.OptimizationDataset("testing", client=client)

    opt_spec = {"program": "geometric"}
    qc_spec = {"driver": "gradient", "method": "UFF", "program": "rdkit"}

    ds.add_specification("test", opt_spec, qc_spec, protocols={"trajectory": "final"})
    ds.add_specification("test2", opt_spec, qc_spec, protocols={"trajectory": "initial_and_final"})

    hooh1 = ptl.data.get_molecule("hooh.json")
    hooh2 = hooh1.copy(update={"geometry": hooh1.geometry + np.array([0, 0, 0.2])})

    ds.add_entry("hooh1", hooh1)
    ds.add_entry("hooh1-2", hooh1)
    ds.add_entry("hooh2", hooh2)

    ds.compute("test")
    ds.compute("test2", subset=["hooh1"])
    fractal_compute_server.await_results()

    ds.query("test")
    ds.query("test2")

    status = ds.status()
    assert status.loc["COMPLETE", "test"] == 3
    assert status.loc["COMPLETE", "test2"] == 1

    status_spec = ds.status(["test", "test2"])
    assert status_spec.loc["COMPLETE", "test"] == 3
    assert status_spec.loc["COMPLETE", "test2"] == 1

    counts = ds.counts()
    assert counts.loc["hooh1", "test"] == 9
    assert np.isnan(counts.loc["hooh2", "test2"])
    assert len(ds.df.loc["hooh1", "test"].trajectory) == 1

    final_energy = 0.00011456853977485626
    for idx, row in ds.df["test"].items():
        assert pytest.approx(row.get_final_energy(), abs=1.0e-5) == final_energy

    opt = ds.get_record("hooh1", "test")
    assert pytest.approx(opt.get_final_energy(), abs=1.0e-5) == final_energy


@testing.using_geometric
@testing.using_rdkit
def test_grid_optimization_dataset(fractal_compute_server):

    client = ptl.FractalClient(fractal_compute_server)

    ds = ptl.collections.GridOptimizationDataset("testing", client=client)

    opt_spec = {"program": "geometric", "keywords": {}}
    qc_spec = {"driver": "gradient", "method": "UFF", "program": "rdkit"}
    ds.add_specification("test", opt_spec, qc_spec)

    hooh1 = ptl.data.get_molecule("hooh.json")

    scans = [{"type": "dihedral", "indices": [0, 1, 2, 3], "steps": [-10, 10], "step_type": "relative"}]
    ds.add_entry("hooh1", hooh1, scans=scans, preoptimization=False)

    ds.compute("test")

    # Test detail status
    fractal_compute_server.await_services(max_iter=1)
    status_detail = ds.status("test", detail=True)
    assert status_detail.loc["hooh1", "Complete Tasks"] == 1

    # Check completions
    fractal_compute_server.await_services()

    ds.query("test")
    assert ds.get_record("hooh1", "test").status == "COMPLETE"


@pytest.mark.parametrize("ds_class", [ptl.collections.Dataset, ptl.collections.ReactionDataset])
def test_get_collection_no_records(ds_class, fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)
    ds = ds_class(f"tnr_{ds_class.__name__}", client=client)
    ds.save()

    ds = client.get_collection(ds_class.__name__, ds.name)
    assert ds.data.records is None


def test_gradient_dataset_lazy_entries_values(gradient_dataset_fixture):
    client, ds = gradient_dataset_fixture

    ds._clear_cache()
    assert ds.data.records is None
    assert ds.data.contributed_values is None

    ds.get_values(subset=["He1"], basis="3-21g")
    assert (~ds.df.isna()).sum().sum() == 1
    if not ds._use_view():
        assert ds.data.records is not None
    else:
        assert ds.data.records is None


def test_get_collection_no_records_ds(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)
    ds = ptl.collections.Dataset("tnr_test", client=client)
    ds.add_entry("He1", ptl.Molecule.from_data("He -1 0 0\n--\nHe 0 0 1"))
    ds.save()

    ds = client.get_collection("dataset", ds.name)
    assert ds.data.records is None
    ds.get_entries()
    assert len(ds.data.records) == 1
    assert ds.data.records[0].name == "He1"


def test_list_collection_group(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)

    ds1 = ptl.collections.Dataset(name="tlco_ds1", client=client)
    ds1.save()
    assert (client.list_collections().reset_index().name == "tlco_ds1").any()

    ds2 = ptl.collections.ReactionDataset(name="tlco_ds2", client=client, group="group1")
    ds2.save()
    assert not (client.list_collections().reset_index().name == "tlco_ds2").any()

    assert (client.list_collections(group=None).reset_index().name == "tlco_ds1").any()
    assert (client.list_collections(group=None).reset_index().name == "tlco_ds2").any()


def test_list_collection_visibility(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)
    ds1 = ptl.collections.GridOptimizationDataset(name="tlcv_ds1", client=client, visibility=False)
    ds2 = ptl.collections.GridOptimizationDataset(name="tlcv_ds2", client=client, visibility=True)
    ds3 = ptl.collections.GridOptimizationDataset(name="tlcv_ds3", client=client)
    ds1.save()
    ds2.save()
    ds3.save()

    names = list(client.list_collections().reset_index().name)
    assert "tlcv_ds1" not in names
    assert "tlcv_ds2" in names
    assert "tlcv_ds3" in names

    names = list(client.list_collections(show_hidden=True).reset_index().name)
    assert "tlcv_ds1" in names
    assert "tlcv_ds2" in names
    assert "tlcv_ds3" in names


def test_collection_metadata(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)

    ds = ptl.collections.Dataset("test_collection_metadata", client=client)
    ds.data.metadata["data_points"] = 133_885
    ds.save()

    assert client.get_collection("dataset", ds.name).data.metadata["data_points"] == 133_885


def test_list_collection_tags(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)

    ptl.collections.Dataset("test_list_collection_tags_1", client=client, tags=[]).save()
    ptl.collections.Dataset("test_list_collection_tags_2", client=client, tags=["t1"]).save()
    ptl.collections.Dataset("test_list_collection_tags_3", client=client, tags=["t2"]).save()
    ptl.collections.Dataset("test_list_collection_tags_4", client=client, tags=["t1", "t2"]).save()

    names = list(client.list_collections().reset_index().name)
    assert "test_list_collection_tags_1" in names
    assert "test_list_collection_tags_2" in names
    assert "test_list_collection_tags_3" in names
    assert "test_list_collection_tags_4" in names

    names = list(client.list_collections(tag="t1").reset_index().name)
    assert "test_list_collection_tags_1" not in names
    assert "test_list_collection_tags_2" in names
    assert "test_list_collection_tags_3" not in names
    assert "test_list_collection_tags_4" in names

    names = list(client.list_collections(tag=["t1"]).reset_index().name)
    assert "test_list_collection_tags_1" not in names
    assert "test_list_collection_tags_2" in names
    assert "test_list_collection_tags_3" not in names
    assert "test_list_collection_tags_4" in names

    names = list(client.list_collections(tag=["t1", "t2"]).reset_index().name)
    assert "test_list_collection_tags_1" not in names
    assert "test_list_collection_tags_2" in names
    assert "test_list_collection_tags_3" in names
    assert "test_list_collection_tags_4" in names


def test_delete_collection(fractal_compute_server):
    client = ptl.FractalClient(fractal_compute_server)
    client.list_collections()

    ds = ptl.collections.Dataset("test_delete_collection", client=client)
    ds.save()

    dsid = ds.data.id

    client.delete_collection("dataset", ds.name)
    assert ds.name.lower() not in client.list_collections(collection_type="dataset", aslist=True)

    # Fails at get_collection
    with pytest.raises(KeyError):
        client.delete_collection("dataset", ds.name)

    # Test DELETE failure specifically
    with pytest.raises(IOError):
        client._automodel_request(f"collection/{dsid}", "delete", payload={"meta": {}}, full_return=True)



"""
Utility functions for on-node procedures.
"""

import json

from typing import Optional, Dict, Any

from qcelemental.models import ResultInput

from ..interface.models import Molecule, QCSpecification


def unpack_single_task_spec(storage, meta, molecules):
    """Transforms a metadata compute packet into an expanded
    QC Schema for multiple runs.

    Parameters
    ----------
    storage : DBSocket
        A live connection to the current database.
    meta : dict
        A JSON description of the metadata involved with the computation
    molecules : list of str, dict
        A list of molecule ID's or full JSON molecules associated with the run.

    Returns
    -------
    ret : tuple(dict, list)
        A dictionary of JSON representations with keys built in.
        The list is an array of any errors occurred

    Examples
    --------

    >>> meta = {
        "procedure": "single",
        "driver": "energy",
        "method": "HF",
        "basis": "sto-3g",
        "keywords": "default",
        "program": "psi4",
    }

    >>> molecules = [{"geometry": [0, 0, 0], "symbols" : ["He"]}]

    >>> unpack_single_task_spec(storage, meta, molecules)

    """

    # Get the required molecules
    raw_molecules_query = storage.get_add_molecules_mixed(molecules)

    # Pull out the needed keywords
    if meta["keywords"] is None:
        keyword_set = {}
    else:
        keyword_set = storage.get_add_keywords_mixed([meta["keywords"]])["data"][0]
        keyword_set = keyword_set["values"]

    # Create the "universal header"
    task_meta = json.dumps(
        {
            "driver": meta["driver"],
            "keywords": keyword_set,
            "model": {"method": meta["method"], "basis": meta["basis"]},
        }
    )

    tasks = []
    for mol in raw_molecules_query["data"]:
        if mol is None:
            tasks.append(None)
            continue

        data = json.loads(task_meta)
        data["molecule"] = mol

        tasks.append(ResultInput(**data))

    return tasks, []


def parse_single_tasks(storage, results, qc_spec):
    """Summary

    Parameters
    ----------
    storage : DBSocket
        A live connection to the current database.
    results : dict
        A (key, result) dictionary of the single return results.

    Returns
    -------

    Examples
    --------

    """

    for k, v in results.items():
        # Flatten data back out
        v["method"] = v["model"]["method"]
        v["basis"] = v["model"]["basis"]
        del v["model"]

        # Molecule should be by ID
        v["molecule"] = storage.add_molecules([Molecule(**v["molecule"])])["data"][0]

        v["keywords"] = qc_spec.keywords
        v["program"] = qc_spec.program

        # Old tags that may still exist if the task was created with previous versions.
        # It is harmless if they do, but may as well do a consistency check
        if "_qcfractal_tags" in v["extras"]:
            assert int(v["extras"]["_qcfractal_tags"]["keywords"]) == int(qc_spec.keywords)
            assert v["extras"]["_qcfractal_tags"]["program"] == qc_spec.program
            del v["extras"]["_qcfractal_tags"]

        del v["schema_name"]
        del v["schema_version"]

        if v.pop("success"):
            v["status"] = "COMPLETE"
        else:
            v["status"] = "ERROR"

    return results


def form_qcinputspec_schema(qc_spec: QCSpecification, keywords: Optional["KeywordSet"] = None) -> Dict[str, Any]:
    if qc_spec.keywords:
        assert keywords.id == qc_spec.keywords

    # Note: program is unused in QCInputSpecification
    ret = {
        "driver": str(qc_spec.driver.name),
        "model": {"method": qc_spec.method},
    }  # yapf: disable
    if qc_spec.basis:
        ret["model"]["basis"] = qc_spec.basis

    if keywords:
        ret["keywords"] = keywords.values
    else:
        ret["keywords"] = {}

    return ret



"""
Tests the imports and exports of the Molecule object.
"""

import json

import numpy as np
import pytest
import qcelemental as qcel

from . import portal as ptl


def test_molecule_constructors():

    ### Water Dimer
    water_psi = ptl.data.get_molecule("water_dimer_minima.psimol")
    ele = np.array([8, 1, 1, 8, 1, 1]).reshape(-1, 1)
    npwater = np.hstack((ele, water_psi.geometry * qcel.constants.conversion_factor("Bohr", "angstrom")))
    water_from_np = ptl.Molecule.from_data(npwater, name="water dimer", dtype="numpy", frags=[3])

    assert water_psi.compare(water_from_np)
    assert water_psi.get_molecular_formula() == "H4O2"

    # Check the JSON construct/deconstruct
    water_from_json = ptl.Molecule(**water_psi.dict())
    assert water_psi.compare(water_from_json)

    ### Neon Tetramer
    neon_from_psi = ptl.data.get_molecule("neon_tetramer.psimol")
    ele = np.array([10, 10, 10, 10]).reshape(-1, 1)
    npneon = np.hstack((ele, neon_from_psi.geometry))
    neon_from_np = ptl.Molecule.from_data(npneon, name="neon tetramer", dtype="numpy", frags=[1, 2, 3], units="bohr")

    assert neon_from_psi.compare(neon_from_np)

    # Check the JSON construct/deconstruct
    neon_from_json = ptl.Molecule(**neon_from_psi.dict())
    assert neon_from_psi.compare(neon_from_json)
    assert neon_from_json.get_molecular_formula() == "Ne4"

    assert water_psi.compare(ptl.Molecule.from_data(water_psi.to_string("psi4")))


def test_water_minima_data():
    mol = ptl.data.get_molecule("water_dimer_minima.psimol")

    assert sum(x == y for x, y in zip(mol.symbols, ["O", "H", "H", "O", "H", "H"])) == mol.geometry.shape[0]
    assert mol.molecular_charge == 0
    assert mol.molecular_multiplicity == 1
    assert np.sum(mol.real) == mol.geometry.shape[0]
    assert np.allclose(mol.fragments, [[0, 1, 2], [3, 4, 5]])
    assert np.allclose(mol.fragment_charges, [0, 0])
    assert np.allclose(mol.fragment_multiplicities, [1, 1])
    assert hasattr(mol, "provenance")
    assert np.allclose(
        mol.geometry,
        [
            [2.81211080, 0.1255717, 0.0],
            [3.48216664, -1.55439981, 0.0],
            [1.00578203, -0.1092573, 0.0],
            [-2.6821528, -0.12325075, 0.0],
            [-3.27523824, 0.81341093, 1.43347255],
            [-3.27523824, 0.81341093, -1.43347255],
        ],
    )  # yapf: disable
    assert mol.get_hash() == "3c4b98f515d64d1adc1648fe1fe1d6789e978d34"


def test_water_minima_fragment():

    mol = ptl.data.get_molecule("water_dimer_minima.psimol")

    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)
    assert frag_0.get_hash() == "5f31757232a9a594c46073082534ca8a6806d367"
    assert frag_1.get_hash() == "bdc1f75bd1b7b999ff24783d7c1673452b91beb9"

    frag_0_1 = mol.get_fragment(0, 1)
    frag_1_0 = mol.get_fragment(1, 0)

    assert np.array_equal(mol.symbols[:3], frag_0.symbols)
    assert np.allclose(mol.masses[:3], frag_0.masses)

    assert np.array_equal(mol.symbols, frag_0_1.symbols)
    assert np.allclose(mol.geometry, frag_0_1.geometry)

    assert np.array_equal(np.hstack((mol.symbols[3:], mol.symbols[:3])), frag_1_0.symbols)
    assert np.allclose(np.hstack((mol.masses[3:], mol.masses[:3])), frag_1_0.masses)


def test_pretty_print():

    mol = ptl.data.get_molecule("water_dimer_minima.psimol")
    assert isinstance(mol.pretty_print(), str)


def test_to_string():

    mol = ptl.data.get_molecule("water_dimer_minima.psimol")
    assert isinstance(mol.to_string("psi4"), str)


def test_water_orient():
    # These are identical molecules, should find the correct results

    mol = ptl.data.get_molecule("water_dimer_stretch.psimol")
    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)

    # Make sure the fragments match
    assert frag_0.get_hash() == frag_1.get_hash()

    # Make sure the complexes match
    frag_0_1 = mol.get_fragment(0, 1, orient=True, group_fragments=True)
    frag_1_0 = mol.get_fragment(1, 0, orient=True, group_fragments=True)

    assert frag_0_1.get_hash() == frag_1_0.get_hash()

    mol = ptl.data.get_molecule("water_dimer_stretch2.psimol")
    frag_0 = mol.get_fragment(0, orient=True)
    frag_1 = mol.get_fragment(1, orient=True)

    # Make sure the fragments match
    assert frag_0.molecular_multiplicity == 1
    assert frag_0.get_hash() == frag_1.get_hash()

    # Make sure the complexes match
    frag_0_1 = mol.get_fragment(0, 1, orient=True, group_fragments=True)
    frag_1_0 = mol.get_fragment(1, 0, orient=True, group_fragments=True)

    # Ghost fragments should prevent overlap
    assert frag_0_1.molecular_multiplicity == 1
    assert frag_0_1.get_hash() != frag_1_0.get_hash()


def test_molecule_errors():
    mol = ptl.data.get_molecule("water_dimer_stretch.psimol")

    data = mol.dict()
    data["whatever"] = 5
    with pytest.raises(ValueError):
        ptl.Molecule(**data)


def test_molecule_repeated_hashing():

    mol = ptl.Molecule(
        **{
            "symbols": ["H", "O", "O", "H"],
            "geometry": [
                1.73178198,
                1.29095807,
                1.03716028,
                1.31566305,
                -0.007440200000000001,
                -0.28074722,
                -1.3143081,
                0.00849608,
                -0.27416914,
                -1.7241109,
                -1.30793432,
                1.02770172,
            ],
        }
    )  # yapf: disable

    h1 = mol.get_hash()
    assert mol.get_molecular_formula() == "H2O2"

    mol2 = ptl.Molecule(**json.loads(mol.json()), orient=False)
    assert h1 == mol2.get_hash()

    mol3 = ptl.Molecule(**json.loads(mol2.json()), orient=False)
    assert h1 == mol3.get_hash()



"""
Tests the interface portal adapter to the REST API
"""

import numpy as np
import pytest

import qcfractal.interface as ptl
from qcfractal.testing import test_server

# All tests should import test_server, but not use it
# Make PyTest aware that this module needs the server

valid_encodings = ["json", "json-ext", "msgpack-ext"]


@pytest.mark.parametrize("encoding", valid_encodings)
def test_client_molecule(test_server, encoding):

    client = ptl.FractalClient(test_server)
    client._set_encoding(encoding)

    water = ptl.data.get_molecule("water_dimer_minima.psimol")
    water.geometry[:] += np.random.random(water.geometry.shape)

    # Test add
    ret = client.add_molecules([water])

    # Test get
    get_mol = client.query_molecules(id=ret[0])
    assert water.compare(get_mol[0])

    # Test molecular_formula get
    get_mol = client.query_molecules(molecular_formula="H4O2")
    assert len(get_mol)


@pytest.mark.parametrize("encoding", valid_encodings)
def test_client_keywords(test_server, encoding):

    client = ptl.FractalClient(test_server)
    client._set_encoding(encoding)

    opt = ptl.models.KeywordSet(values={"one": "fish", "two": encoding})

    # Test add
    ret = client.add_keywords([opt])

    # Test get
    get_kw = client.query_keywords([ret[0]])
    assert opt == get_kw[0]

    get_kw = client.query_keywords(hash_index=[opt.hash_index])
    assert opt == get_kw[0]


@pytest.mark.parametrize("encoding", valid_encodings)
def test_client_duplicate_keywords(test_server, encoding):

    client = ptl.FractalClient(test_server)
    client._set_encoding(encoding)

    key_name = f"key-{encoding}"
    opt1 = ptl.models.KeywordSet(values={key_name: 1})
    opt2 = ptl.models.KeywordSet(values={key_name: 2})
    opt3 = ptl.models.KeywordSet(values={key_name: 3})

    # Test add
    ret = client.add_keywords([opt1, opt1])
    assert len(ret) == 2
    assert ret[0] == ret[1]

    ret2 = client.add_keywords([opt1])
    assert len(ret2) == 1
    assert ret2[0] == ret[0]

    ret3 = client.add_keywords([opt2, opt1, opt3])
    assert len(ret3) == 3
    assert ret3[1] == ret[0]


@pytest.mark.parametrize("encoding", valid_encodings)
def test_empty_query(test_server, encoding):

    client = ptl.FractalClient(test_server)
    client._set_encoding(encoding)

    with pytest.raises(IOError) as error:
        client.query_procedures(limit=1)

    assert "ID is required" in str(error.value)


@pytest.mark.parametrize("encoding", valid_encodings)
def test_collection_portal(test_server, encoding):

    db_name = f"Torsion123-{encoding}"
    db = {
        "collection": "torsiondrive",
        "name": db_name,
        "something": "else",
        "array": ["12345"],
        "visibility": True,
        "view_available": False,
        "group": "default",
    }

    client = ptl.FractalClient(test_server)
    client._set_encoding(encoding)

    # Test add
    ret = client.add_collection(db, full_return=True)
    print(ret)

    # Test get
    get_db = client.get_collection(db["collection"], db["name"], full_return=True)
    db_id = get_db.data[0].pop("id")

    # got a default values when created
    get_db.data[0].pop("tags", None)
    get_db.data[0].pop("tagline", None)
    get_db.data[0].pop("provenance", None)
    get_db.data[0].pop("view_url_hdf5", None)
    get_db.data[0].pop("view_url_plaintext", None)
    get_db.data[0].pop("view_metadata", None)
    get_db.data[0].pop("description", None)

    assert db == get_db.data[0]

    # Test add w/o overwrite
    ret = client.add_collection(db, full_return=True)
    assert ret.meta.success is False

    # Test that client is smart enough to trap non-id'ed overwrites
    with pytest.raises(KeyError):
        _ = client.add_collection(db, overwrite=True)

    # Test that we cannot use a local key
    db["id"] = "local"
    db["array"] = ["6789"]
    with pytest.raises(KeyError):
        _ = client.add_collection(db, overwrite=True)

    # Finally test that we can overwrite
    db["id"] = db_id
    r = client.add_collection(db, overwrite=True)
    get_db = client.get_collection(db["collection"], db["name"], full_return=True)
    assert get_db.data[0]["array"] == ["6789"]


@pytest.mark.parametrize("encoding", valid_encodings)
def test_custom_queries(test_server, encoding):
    """ Test the round trip between client and server in custom queries"""

    client = ptl.FractalClient(test_server)
    client._set_encoding(encoding)

    # Dummy test, not found
    ret = client.custom_query("optimization", "final_result", {"optimization_ids": [1]}, full_return=True)

    assert ret.meta.success
    assert ret.meta.n_found == 0



"""Remove error field from task_queue

Revision ID: 5686f64a46c8
Revises: dd35678aae90
Create Date: 2020-10-27 14:06:58.794701

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "5686f64a46c8"
down_revision = "dd35678aae90"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("task_queue", "error")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("task_queue", sa.Column("error", sa.VARCHAR(), autoincrement=False, nullable=True))
    # ### end Alembic commands ###



"""
Tests for visualization code
"""

import pytest
import requests

from . import portal

try:
    import plotly

    _has_ploty = True
except ModuleNotFoundError:
    _has_ploty = False

using_plotly = pytest.mark.skipif(
    _has_ploty is False, reason="Not detecting module 'plotly'. Install package if necessary to enable tests."
)


def live_fractal_or_skip():
    """
    Ensure Fractal live connection can be made
    First looks for a local staging server, then tries QCArchive.
    """
    try:
        return portal.FractalClient("localhost:7777", verify=False)
    except (requests.exceptions.ConnectionError, ConnectionRefusedError):
        print("Failed to connect to localhost")
        try:
            requests.get("https://api.qcarchive.molssi.org:443", json={}, timeout=5)
            return portal.FractalClient()
        except (requests.exceptions.ConnectionError, ConnectionRefusedError):
            return pytest.skip("Could not make a connection to central Fractal server")


@pytest.fixture
def S22Fixture():

    # Connect to the primary database
    client = live_fractal_or_skip()

    S22 = client.get_collection("ReactionDataset", "S22")

    return client, S22


@using_plotly
@pytest.mark.parametrize("kind", ["bar", "violin"])
def test_plot_dataset(S22Fixture, kind):

    client, S22 = S22Fixture

    fig = S22.visualize(
        method=["b2plyp", "pbe"], basis=["def2-svp", "def2-TZVP"], return_figure=True, bench="S22a", kind=kind
    ).to_dict()
    assert "S22" in fig["layout"]["title"]["text"]


@using_plotly
@pytest.mark.parametrize("kind", ["violin", "bar"])
@pytest.mark.parametrize("groupby", ["method", "basis"])
def test_plot_dataset_groupby(S22Fixture, kind, groupby):

    client, S22 = S22Fixture

    fig = S22.visualize(
        method=["b2plyp", "b3lyp"],
        basis=["def2-svp", "def2-TZVP"],
        return_figure=True,
        bench="S22a",
        kind=kind,
        groupby=groupby,
    ).to_dict()
    assert "S22" in fig["layout"]["title"]["text"]


@using_plotly
def test_plot_qca_examples(S22Fixture):
    """ Tests plotting examples from QCArchiveExamples/basic_examples/reaction_dataset.ipynb"""
    client, S22 = S22Fixture
    fig = S22.visualize(method=["B3LYP", "B3LYP-D3", "B3LYP-D3M"], basis=["def2-tzvp"], groupby="D3").to_dict()
    assert "S22" in fig["layout"]["title"]["text"]
    fig = S22.visualize(
        method=["B3LYP", "B3LYP-D3", "B2PLYP", "B2PLYP-D3"], basis="def2-tzvp", groupby="D3", kind="violin"
    )
    assert "S22" in fig["layout"]["title"]["text"]


### Test TorsionDriveDataset scans


@pytest.fixture
def TDDSFixture():
    pytest.skip("Database and tests are out of sync, fixed in 6.1 release.")

    # Connect to the primary database
    client = live_fractal_or_skip()
    TDDs = client.get_collection("TorsionDriveDataset", "OpenFF Fragmenter Phenyl Benchmark")

    return client, TDDs


@using_plotly
def test_plot_torsiondrive_dataset(TDDSFixture):
    client, ds = TDDSFixture

    ds.visualize("[CH3:4][O:3][c:2]1[cH:1]cccc1", ["B3LYP-D3", "UFF"], units="kJ / mol", return_figure=True)
    ds.visualize(
        ["[CH3:4][O:3][c:2]1[cH:1]cccc1", "[CH3:4][O:3][c:2]1[cH:1]ccnc1"], "UFF", relative=False, return_figure=True
    )


@using_plotly
def test_plot_torsiondrive_dataset_measured(TDDSFixture):
    client, ds = TDDSFixture

    ds.visualize(
        "[CH3:4][O:3][c:2]1[cH:1]cccc1", "B3LYP-D3", units="kJ / mol", use_measured_angle=True, return_figure=True
    )



import pydantic
import pytest

from ..rest_models import QueryFilter


def test_include_exclude_exclusive():
    QueryFilter()
    QueryFilter(include=["foo"])
    QueryFilter(exclude=["goo"])
    with pytest.raises(pydantic.ValidationError):
        QueryFilter(include=["foo"], exclude=["goo"])



"""
QCPortal Database ODM
"""
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set

from ..models import GridOptimizationInput, ObjectId, OptimizationSpecification, ProtoModel, QCSpecification
from ..models.gridoptimization import GOKeywords
from .collection import BaseProcedureDataset
from .collection_utils import register_collection

if TYPE_CHECKING:  # pragma: no cover
    from ..models.gridoptimization import ScanDimension
    from ..models import Molecule


class GOEntry(ProtoModel):
    """Data model for the `reactions` list in Dataset"""

    name: str
    initial_molecule: ObjectId
    go_keywords: GOKeywords
    attributes: Dict[str, Any]  # Might be overloaded key types
    object_map: Dict[str, ObjectId] = {}


class GOEntrySpecification(ProtoModel):
    name: str
    description: Optional[str]
    optimization_spec: OptimizationSpecification
    qc_spec: QCSpecification


class GridOptimizationDataset(BaseProcedureDataset):
    class DataModel(BaseProcedureDataset.DataModel):

        records: Dict[str, GOEntry] = {}
        history: Set[str] = set()
        specs: Dict[str, GOEntrySpecification] = {}

        class Config(BaseProcedureDataset.DataModel.Config):
            pass

    def _internal_compute_add(self, spec: Any, entry: Any, tag: str, priority: str) -> ObjectId:
        service = GridOptimizationInput(
            initial_molecule=entry.initial_molecule,
            keywords=entry.go_keywords,
            optimization_spec=spec.optimization_spec,
            qc_spec=spec.qc_spec,
        )

        return self.client.add_service([service], tag=tag, priority=priority).ids[0]

    def add_specification(
        self,
        name: str,
        optimization_spec: OptimizationSpecification,
        qc_spec: QCSpecification,
        description: str = None,
        overwrite=False,
    ) -> None:
        """
        Parameters
        ----------
        name : str
            The name of the specification
        optimization_spec : OptimizationSpecification
            A full optimization specification for TorsionDrive
        qc_spec : QCSpecification
            A full quantum chemistry specification for TorsionDrive
        description : str, optional
            A short text description of the specification
        overwrite : bool, optional
            Overwrite existing specification names

        """

        spec = GOEntrySpecification(
            name=name, optimization_spec=optimization_spec, qc_spec=qc_spec, description=description
        )

        return self._add_specification(name, spec, overwrite=overwrite)

    def add_entry(
        self,
        name: str,
        initial_molecule: "Molecule",
        scans: List["ScanDimension"],
        preoptimization: bool = True,
        attributes: Dict[str, Any] = None,
        save: bool = True,
    ) -> None:
        """
        Parameters
        ----------
        name : str
            The name of the entry, will be used for the index
        initial_molecule : Molecule
            The initial molecule to start the GridOptimization
        scans : List[ScanDimension]
            A list of ScanDimension objects detailing the dimensions to scan over.
        preoptimization : bool, optional
            If True, pre-optimizes the molecules before scanning, otherwise
        attributes : Dict[str, Any], optional
            Additional attributes and descriptions for the entry
        save : bool, optional
            If true, saves the collection after adding the entry. If this is False be careful
            to call save after all entries are added, otherwise data pointers may be lost.

        """

        self._check_entry_exists(name)  # Fast skip

        if attributes is None:
            attributes = {}

        # Build new objects
        molecule_id = self.client.add_molecules([initial_molecule])[0]
        go_keywords = GOKeywords(scans=scans, preoptimization=preoptimization)

        record = GOEntry(name=name, initial_molecule=molecule_id, go_keywords=go_keywords, attributes=attributes)
        self._add_entry(name, record, save)


register_collection(GridOptimizationDataset)



"""
Importer for the DB socket class.
"""

__all__ = ["storage_socket_factory", "ViewHandler"]

from .db_queries import TorsionDriveQueries
from .storage_socket import storage_socket_factory
from .view import ViewHandler



import atexit
import os
import shutil
import subprocess
import tempfile
import time
import re
from typing import Any, Dict, List, Optional, Union

import psycopg2
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from qcfractal.storage_sockets.models import Base, VersionsORM

from .config import FractalConfig
from .port_util import find_port, is_port_open


class PostgresHarness:
    def __init__(self, config: Union[Dict[str, Any], FractalConfig], quiet: bool = True, logger: "print" = print):
        """A flexible connection to a PostgreSQL server

        Parameters
        ----------
        config : Union[Dict[str, Any], FractalConfig]
            The configuration options
        quiet : bool, optional
            If True, does not log any operations
        logger : print, optional
            The logger to show the operations to.
        """
        if isinstance(config, dict):
            config = FractalConfig(**config)
        self.config = config
        self.quiet = quiet
        self._logger = logger
        self._checked = False
        self._alembic_ini = os.path.join(os.path.abspath(os.path.dirname(__file__)), "alembic.ini")

    def _run(self, commands):
        proc = subprocess.run(commands, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout = proc.stdout.decode()
        if not self.quiet:
            self.logger(stdout)

        ret = {"retcode": proc.returncode, "stdout": stdout, "stderr": proc.stderr.decode()}

        return ret

    def _check_psql(self) -> None:
        """
        Checks to see if the proper PostgreSQL commands are present. Raises a ValueError if they are not found.
        """

        if self.config.database.host != "localhost":
            raise ValueError(f"Cannot modify PostgreSQL as configuration points to non-localhost: {self.config.host}")

        if self._checked:
            return

        msg = """
Could not find 'pg_ctl' in the current path. Please install PostgreSQL with 'conda install postgresql'.

Alternatively, you can install a system PostgreSQL manually, please see the following link: https://www.postgresql.org/download/
"""

        if shutil.which("pg_ctl") is None:
            raise ValueError(msg)
        else:
            self._checked = True

    def logger(self, msg: str) -> None:
        """Prints a logging message depending on quiet settings.

        Parameters
        ----------
        msg : str
            The message to show.
        """
        if not self.quiet:
            self._logger(msg)

    def database_uri(self) -> str:
        """Provides the full PostgreSQL URI string.

        Returns
        -------
        str
            The database URI
        """
        return self.config.database_uri(safe=False, database="")

    def connect(self, database: Optional[str] = None) -> "Connection":
        """Builds a psycopg2 connection object.

        Parameters
        ----------
        database : Optional[str], optional
            The database to connect to, otherwise defaults to None

        Returns
        -------
        Connection
            A live Connection object.
        """
        if database is None:
            database = "postgres"
        return psycopg2.connect(
            database=database,
            user=self.config.database.username,
            host=self.config.database.host,
            port=self.config.database.port,
            password=self.config.database.password,
        )

    def is_alive(self, database: Optional[str] = None) -> bool:
        """Checks if the postgres is alive, and optionally if the database is present.

        Parameters
        ----------
        database : Optional[str], optional
            The datbase to connect to

        Returns
        -------
        bool
            If True, the postgres database is alive.
        """
        try:
            self.connect(database=database)
            return True
        except psycopg2._psycopg.OperationalError:
            return False

    def command(self, cmd: str, check: bool = True) -> Any:
        """Runs psql commands and returns their output while connected to the correct postgres instance.

        Parameters
        ----------
        cmd : str
            A psql command string.
            Description

        """
        self._check_psql()

        if isinstance(cmd, str):
            cmd = [cmd]

        self.logger(f"pqsl command: {cmd}")
        psql_cmd = [shutil.which("psql"), "-p", str(self.config.database.port), "-c"]

        cmd = self._run(psql_cmd + cmd)
        if check:
            if cmd["retcode"] != 0:
                raise ValueError("psql operation did not complete.")
        return cmd

    def pg_ctl(self, cmds: List[str]) -> Any:
        """Runs pg_ctl commands and returns their output while connected to the correct postgres instance.

        Parameters
        ----------
        cmds : List[str]
            A list of PostgreSQL pg_ctl commands to run.
        """
        self._check_psql()

        self.logger(f"pg_ctl command: {cmds}")
        psql_cmd = [shutil.which("pg_ctl"), "-D", str(self.config.database_path)]
        return self._run(psql_cmd + cmds)

    def create_database(self, database_name: str) -> bool:
        """Creates a new database for the current postgres instance. If the database is existing, no
        changes to the database are made.

        Parameters
        ----------
        database_name : str
            The name of the database to create.

        Returns
        -------
        bool
            If the operation was successful or not.
        """
        conn = self.connect()
        conn.autocommit = True

        cursor = conn.cursor()
        cursor.execute(f"SELECT 1 FROM pg_catalog.pg_database WHERE datname = '{database_name}'")
        exists = cursor.fetchone()

        if not exists:
            cursor.execute(f"CREATE DATABASE {database_name}")

        return self.is_alive(database=database_name)

    def create_tables(self):
        """Create database tables using SQLAlchemy models"""

        uri = self.config.database_uri(safe=False)
        self.logger(f"Creating tables for database: {uri}")
        engine = create_engine(uri, echo=False, pool_size=1)

        # actually create the tables
        try:
            Base.metadata.create_all(engine)
        except Exception as e:
            raise ValueError(f"SQLAlchemy Connection Error\n {str(e)}")

        return True

    def update_db_version(self):
        """Update current version of QCFractal in the DB"""

        uri = self.config.database_uri(safe=False)

        engine = create_engine(uri, echo=False, pool_size=1)
        session = sessionmaker(bind=engine)()
        try:
            import qcfractal, qcelemental, qcengine

            elemental_version = qcelemental.__version__
            fractal_version = qcfractal.__version__
            engine_version = qcengine.__version__

            self.logger(f"Updating current version of QCFractal in DB: {uri} \n" f"to version {qcfractal.__version__}")
            current_ver = VersionsORM(
                elemental_version=elemental_version, fractal_version=fractal_version, engine_version=engine_version
            )
            session.add(current_ver)
            session.commit()
        except Exception as e:
            raise ValueError(f"Failed to Update DB version.\n {str(e)}")
        finally:
            session.close()

        return True

    def upgrade(self):
        """
        Upgrade the database schema using the latest alembic revision.
        The database data won't be deleted.
        """

        ret = self._run(self.alembic_commands() + ["upgrade", "head"])

        if ret["retcode"] != 0:
            self.logger(ret["stderr"])
            raise ValueError(
                f"\nFailed to Upgrade the database, make sure to init the database first before being able to upgrade it.\n"
            )

        return True

    def start(self) -> Any:
        """
        Starts a PostgreSQL server based off the current configuration parameters. The server must be initialized
        and the configured port open.
        """

        self._check_psql()

        # Startup the server
        self.logger("Starting the database:")

        if is_port_open(self.config.database.host, self.config.database.port):
            self.logger("Service currently running the configured port, current_status:\n")
            status = self.pg_ctl(["status"])

            # If status is ok, exit is 0
            if status["retcode"] != 0:
                raise ValueError(
                    f"A process is already running on 'port:{self.config.database.port}` that is not associated with the PostgreSQL instance at `location:{self.config.database.directory}.`"
                    "\nThis often happens when two PostgreSQL databases are attempted to run on the same port."
                    "\nEither shut down the other PostgreSQL database or change the `qcfractal-server init --db-port`."
                    "\nStopping."
                )

            if not self.is_alive():
                raise ValueError(f"PostgreSQL is running, but cannot connect to the default port.")

            self.logger("Found running PostgreSQL instance with correct configuration.")

        else:
            start_status = self.pg_ctl(
                ["-l", str(self.config.database_path / self.config.database.logfile), "start"]
            )  # yapf: disable

            if not (("server started" in start_status["stdout"]) or ("server starting" in start_status["stdout"])):
                with open(str(self.config.database_path / self.config.database.logfile)) as log_f:
                    log_contents = log_f.read()
                raise ValueError(
                    f"Could not start the PostgreSQL server. Error below:\n\n{start_status['stderr']}\n\nLog contents:\n\n{log_contents}"
                )

            # Check that we are alive
            for x in range(10):
                if self.is_alive():
                    break
                else:
                    time.sleep(0.1)
            else:
                raise ValueError(
                    f"Could not connect to the server after booting. Boot log:\n\n{start_status['stderr']}"
                )

            self.logger("PostgreSQL successfully started in a background process, current_status:\n")
            if not self.quiet:
                self._run([shutil.which("pg_ctl"), "-D", str(self.config.database_path), "status"])  # yapf: disable

        return True

    def shutdown(self) -> Any:
        """Shutsdown the current postgres instance."""

        self._check_psql()

        ret = self.pg_ctl(["stop"])
        return ret

    def initialize_postgres(self) -> None:
        """Initializes and starts the current postgres instance."""

        self._check_psql()

        self.logger("Initializing the Postgresql database:")

        # Initialize the database
        init_status = self._run([shutil.which("initdb"), "-D", self.config.database_path])
        if "Success." not in init_status["stdout"]:
            raise ValueError(f"Could not initialize the PostgreSQL server. Error below:\n\n{init_status['stderr']}")

        # Change any configurations
        psql_conf_file = self.config.database_path / "postgresql.conf"
        psql_conf = psql_conf_file.read_text()
        if self.config.database.port != 5432:
            assert "#port = 5432" in psql_conf
            psql_conf = psql_conf.replace("#port = 5432", f"port = {self.config.database.port}")

            # Change the location of the socket file
            # Some OSs/Linux distributions will use a directory not writeable by a normal user
            psql_conf = re.sub(
                r"#?unix_socket_directories =.*",
                f"unix_socket_directories = '{self.config.database_path}'",
                psql_conf,
                re.M,
            )

            psql_conf_file.write_text(psql_conf)

        # Start the database
        self.start()

        # Create the user and database
        self.logger(f"Building user information.")
        self._run([shutil.which("createdb"), "-p", str(self.config.database.port)])

        success = self.create_database(self.config.database.database_name)

        if success is False:
            self.shutdown()
            raise ValueError("Database created successfully, but could not connect. Shutting down postgres.")

        self.logger("\nDatabase server successfully started!")

    def alembic_commands(self) -> List[str]:
        db_uri = self.config.database_uri(safe=False)
        return [shutil.which("alembic"), "-c", self._alembic_ini, "-x", "uri=" + db_uri]

    def init_database(self) -> None:

        # TODO: drop tables

        # create models
        self.create_tables()

        # update alembic_version table with the current version
        self.logger(f"\nStamping Database with current version..")

        ret = self._run(self.alembic_commands() + ["stamp", "head"])

        if ret["retcode"] != 0:
            self.logger(ret)
            raise ValueError("\nFailed to Stamp the database with current version.\n")

    def backup_database(self, filename: Optional[str] = None) -> None:

        # Reasonable check here
        self._check_psql()

        if filename is None:
            filename = f"{self.config.database.database_name}.bak"

        filename = os.path.realpath(filename)

        # fmt: off
        cmds = [
            shutil.which("pg_dump"),
            "-p", str(self.config.database.port),
            "-d", self.config.database.database_name,
            "-Fc", # Custom postgres format, fast
            "--file", filename
        ]
        # fmt: on

        self.logger(f"pg_backup command: {'  '.join(cmds)}")
        ret = self._run(cmds)

        if ret["retcode"] != 0:
            self.logger(ret)
            raise ValueError("\nFailed to backup the database.\n")

    def restore_database(self, filename) -> None:

        # Reasonable check here
        self._check_psql()

        self.create_database(self.config.database.database_name)

        # fmt: off
        cmds = [
            shutil.which("pg_restore"),
            f"--port={self.config.database.port}",
            f"--dbname={self.config.database.database_name}",
            filename
        ]
        # fmt: on

        self.logger(f"pg_backup command: {'  '.join(cmds)}")
        ret = self._run(cmds)

        if ret["retcode"] != 0:
            self.logger(ret["stderr"])
            raise ValueError("\nFailed to restore the database.\n")

    def database_size(self) -> str:
        """
        Returns a pretty formatted string of the database size.
        """

        return self.command(
            [f"SELECT pg_size_pretty( pg_database_size('{self.config.database.database_name}') );", "-t", "-A"]
        )


class TemporaryPostgres:
    def __init__(
        self,
        database_name: Optional[str] = None,
        tmpdir: Optional[str] = None,
        quiet: bool = True,
        logger: "print" = print,
    ):
        """A PostgreSQL instance run in a temporary folder.

        ! Warning ! All data is lost when this object is deleted.

        Parameters
        ----------
        database_name : Optional[str], optional
            The database name to create.
        tmpdir : Optional[str], optional
            A directory to create the postgres instance in, if not None the data is not deleted upon shutdown.
        quiet : bool, optional
            If True, does not log any operations
        logger : print, optional
            The logger to show the operations to.
        """

        self._active = True

        if not tmpdir:
            self._db_tmpdir = tempfile.TemporaryDirectory()
        else:
            self._db_tmpdir = tmpdir

        self.quiet = quiet
        self.logger = logger

        config_data = {"port": find_port(), "directory": self._db_tmpdir.name}
        if database_name:
            config_data["database_name"] = database_name
        self.config = FractalConfig(database=config_data)
        self.psql = PostgresHarness(self.config)
        self.psql.initialize_postgres()
        self.psql.init_database()

        atexit.register(self.stop)

    def __del__(self):
        """
        Cleans up the TemporaryPostgres instance on delete.
        """

        self.stop()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()
        return False

    def database_uri(self, safe: bool = True, database: Optional[str] = None) -> str:
        """Provides the full Postgres URI string.

        Parameters
        ----------
        safe : bool, optional
            If True, hides the postgres password.
        database : Optional[str], optional
            An optional database to add to the string.

        Returns
        -------
        str
            The database URI
        """
        return self.config.database_uri(safe=safe, database=database)

    def stop(self) -> None:
        """
        Shuts down the Snowflake instance. This instance is not recoverable after a stop call.
        """

        if not self._active:
            return

        self.psql.shutdown()

        # Closed down
        self._active = False
        atexit.unregister(self.stop)


# createuser [-p 5433] --superuser postgres
# psql [-p 5433] -c "create database qcarchivedb;" -U postgres
# psql [-p 5433] -c "create user qcarchive with password 'mypass';" -U postgres
# psql [-p 5433] -c "grant all privileges on database qcarchivedb to qcarchive;" -U postgres



"""
Queue adapter for Fireworks
"""

import logging
from typing import Any, Dict, Hashable, Optional, Tuple

from qcelemental.models import FailedOperation, Optimization, Result
from qcelemental.models.common_models import qcschema_optimization_output_default, qcschema_output_default

from .base_adapter import BaseAdapter

__all__ = ["FireworksAdapter"]

# This, the __all__, and the imports from qcelemental.models are all to make it so Fireworks returns either the
# Result, the Optimization, or the FailedOperation objects like all other adapters instead of a dict.
# In the future, this may be cleaned up but for now, it works
schema_mapper = {qcschema_output_default: Result, qcschema_optimization_output_default: Optimization}


class FireworksAdapter(BaseAdapter):
    def __init__(self, client: Any, logger: Optional[logging.Logger] = None, **kwargs):
        BaseAdapter.__init__(self, client, logger, **kwargs)
        self.client.reset(None, require_password=False, max_reset_wo_password=int(1e8))

    def __repr__(self):
        return "<FireworksAdapter client=<LaunchPad host='{}' name='{}'>>".format(self.client.host, self.client.name)

    def _submit_task(self, task_spec: Dict[str, Any]) -> Tuple[Hashable, Any]:
        import fireworks

        kwargs = task_spec["spec"]["kwargs"]
        kwargs["return_dict"] = True
        fw = fireworks.Firework(
            fireworks.PyTask(
                func=task_spec["spec"]["function"],
                args=task_spec["spec"]["args"],
                kwargs=kwargs,
                stored_data_varname="fw_results",
            ),
            spec={"_launch_dir": "/tmp/"},
        )
        launches = self.client.add_wf(fw)

        return list(launches.values())[0], task_spec["id"]

    def _task_exists(self, lookup):
        """Overload existing method"""
        return False

    def acquire_complete(self) -> Dict[str, Any]:
        ret = {}

        # Pull out completed results that match our queue ids
        cursor = self.client.launches.find(
            {"fw_id": {"$in": list(self.queue.keys())}, "state": {"$in": ["COMPLETED", "FIZZLED"]}},
            {
                "action.stored_data.fw_results": True,
                "action.stored_data._task.args": True,
                "action.stored_data._exception": True,
                "_id": False,
                "fw_id": True,
                "state": True,
            },
        )

        for tmp_data in cursor:
            key = self.queue.pop(tmp_data["fw_id"])
            if tmp_data["state"] == "COMPLETED":
                key_data = tmp_data["action"]["stored_data"]["fw_results"]
                if key_data["success"]:
                    # Cast dict to the Result or Optimization based on the schema
                    ret[key] = schema_mapper[key_data["schema_name"]](**key_data)
                else:
                    ret[key] = FailedOperation(**key_data)
            else:
                blob = tmp_data["action"]["stored_data"]["_task"]["args"][0]
                msg = tmp_data["action"]["stored_data"]["_exception"]["_stacktrace"]
                blob["error_message"] = msg
                blob["success"] = False
                key_data = blob
                ret[key] = FailedOperation(**key_data)

        return ret

    def await_results(self) -> bool:
        # Launch all results consecutively
        import fireworks.core.rocket_launcher

        fireworks.core.rocket_launcher.rapidfire(self.client, strm_lvl="CRITICAL")
        return True

    def close(self) -> bool:
        self.client.reset(None, require_password=False, max_reset_wo_password=int(1e8))
        return True



"""
Copywrite:
This product includes GeoLite2 data created by MaxMind, available from
https://www.maxmind.com
(attribution requirement)
"""

import logging

logger = logging.getLogger(__name__)


class API_AccessLogger:
    """
    Extract access information from HTTP requests to be saved by the database
    Calculate geo data using geoip2 if the library and its files are available
    otherwise, just extracts the basic information
    """

    def __init__(self, geo_file_path):

        self.geoip2_reader = None
        try:
            import geoip2.database

            self.geoip2_reader = geoip2.database.Reader(geo_file_path)
            logger.info(f"Initialized geoip2 with {geo_file_path} successfully.")
        except ImportError:
            logger.error(
                f"Cannot import geoip2 module. To use API access logging, you need "
                f"to install it manually using `pip install geoip2`"
            )
        except FileNotFoundError:
            logger.error(
                f"Geoip cites file cannot be read from {geo_file_path}.\n"
                f"Make sure to manually download the file from: \n"
                f"https://geolite.maxmind.com/download/geoip/database/GeoLite2-City.tar.gz\n"
                f"Then, set the geo_file_path in qcfractal_config.yaml in your base_folder "
                f"(default base_folder is ~/.qca/qcfractal/qcfractal_config.yaml)."
            )

    def get_api_access_log(self, request, access_type=None, extra_params=None):

        log = {}

        if not access_type:
            log["access_type"] = request.uri[1:]  # remove /
        else:
            log["access_type"] = access_type

        log["access_method"] = request.method  # GET or POST

        # get the real IP address behind a proxy or ngnix
        x_real_ip = request.headers.get("X-Real-IP", None)
        log["ip_address"] = x_real_ip or request.remote_ip

        log["user_agent"] = request.headers["User-Agent"]

        # TODO: extract needed info, maybe handle by caller
        log["extra_params"] = extra_params
        # Or, will saved as string anyway
        # log.extra_access_params = request.json

        # extra geo data if available
        extra = self.get_geoip2_data(log["ip_address"])
        log.update(extra)

        return log

    def get_geoip2_data(self, ip_address):
        out = {}

        if not self.geoip2_reader:
            return out

        try:
            loc_data = self.geoip2_reader.city(ip_address)
            out["city"] = loc_data.city.name
            out["country"] = loc_data.country.name
            out["country_code"] = loc_data.country.iso_code
            out["ip_lat"] = str(loc_data.location.latitude)
            out["ip_long"] = str(loc_data.location.longitude)
            out["postal_code"] = loc_data.postal.code
            out["subdivision"] = loc_data.subdivisions.most_specific.name
        except:  # lgtm [py/catch-base-exception]
            logger.error(f"Problem getting geoip2 data for {ip_address}")

        return out



"""
A factory for Database sockets
"""


def storage_socket_factory(uri, project_name="", logger=None, db_type=None, **kwargs):
    """
    Factory for generating storage sockets. Spins up a given storage layer on request given common inputs.

    Right now only supports MongoDB.

    Parameters
    ----------
    uri : string
        A URI to given database such as ("postgresql://localhost:5432", )
    project_name : string
        Name of the project
    logger : logging.Logger, Optional, Default: None
        Specific logger to report to
    db_type : string, Optional, Default: 'sqlalchemy'
        socket type, 'sqlalchemy'
    **kwargs
        Additional keyword arguments to pass to the storage constructor

    Returns
    -------

    """

    if db_type is None:
        # try to find db_type from uri
        if uri.startswith("postgresql"):
            db_type = "sqlalchemy"
        else:
            raise TypeError("Unknown DB type, uri: {}".format(uri))

    if db_type == "sqlalchemy":
        from . import sqlalchemy_socket

        return sqlalchemy_socket.SQLAlchemySocket(uri, project=project_name, logger=logger, **kwargs)
    else:
        raise KeyError("DBType {} not understood".format(db_type))



"""
Helpers for compressing data to send back to the server
"""

from typing import Union, Optional, Dict
from ..interface.models import KVStore, CompressionEnum
from qcelemental.models import AtomicResult, OptimizationResult


def _compress_common(
    result: Union[AtomicResult, OptimizationResult],
    compression: CompressionEnum = CompressionEnum.lzma,
    compression_level: int = None,
):
    """
    Compresses outputs of an AtomicResult or OptimizationResult, storing them in extras
    """

    stdout = result.stdout
    stderr = result.stderr
    error = result.error

    extras = result.extras
    update = {}
    if stdout is not None:
        extras["_qcfractal_compressed_stdout"] = KVStore.compress(stdout, compression, compression_level)
        update["stdout"] = None
    if stderr is not None:
        extras["_qcfractal_compressed_stderr"] = KVStore.compress(stderr, compression, compression_level)
        update["stderr"] = None
    if error is not None:
        extras["_qcfractal_compressed_error"] = KVStore.compress(error, compression, compression_level)
        update["error"] = None

    update["extras"] = extras
    return result.copy(update=update)


def _compress_optimizationresult(
    result: OptimizationResult,
    compression: CompressionEnum = CompressionEnum.lzma,
    compression_level: Optional[int] = None,
):
    """
    Compresses outputs inside an OptimizationResult, storing them in extras

    Outputs for the AtomicResults stored in the trajectory will be stored in the extras for that AtomicResult
    """

    # Handle the trajectory
    trajectory = [_compress_common(x, compression, compression_level) for x in result.trajectory]
    result = result.copy(update={"trajectory": trajectory})

    # Now handle the outputs of the optimization itself
    return _compress_common(result, compression, compression_level)


def compress_results(
    results: Dict[str, Union[AtomicResult, OptimizationResult]],
    compression: CompressionEnum = CompressionEnum.lzma,
    compression_level: int = None,
):
    """
    Compress outputs inside results, storing them in extras

    The compressed outputs are stored in extras. For OptimizationResult, the outputs for the optimization
    are stored in the extras field of the OptimizationResult, while the outputs for the trajectory
    are stored in the extras field for the AtomicResults within the trajectory
    """

    ret = {}
    for k, result in results.items():
        if isinstance(result, AtomicResult):
            ret[k] = _compress_common(result, compression, compression_level)
        elif isinstance(result, OptimizationResult):
            ret[k] = _compress_optimizationresult(result, compression, compression_level)
        else:
            ret[k] = result

    return ret



import io
import pathlib
from typing import Any, Dict, Union

import numpy as np
import pandas as pd
from qcelemental.util.serialization import serialize

from ..interface.collections import HDF5View


class ViewHandler:
    def __init__(self, path: Union[str, pathlib.Path]) -> None:
        """
        Parameters
        ----------
        path: Union[str, Path]
            Directory containing dataset views
        """
        self._view_cache: Dict[int, HDF5View] = {}
        self._path = pathlib.Path(path)
        if not self._path.is_dir():
            raise ValueError(f"Path in ViewHandler must be a directory, got: {self._path}")

    def view_path(self, collection_id: int) -> pathlib.Path:
        """
        Returns the path to a view corresponding to a collection identified by an id.

        Parameters
        ----------
        collection_id: int
            Collection id corresponding to view

        Returns
        -------
        pathlib.Path
            Path of requested view
        """

        return self._path / f"{collection_id}.hdf5"

    def view_exists(self, collection_id: int) -> bool:
        """
        Checks if view corresponding to a collection exists.

        Parameters
        ----------
        collection_id: int
            Collection id corresponding to view

        Returns
        -------
        bool
            Does the view exist?
        """

        return self.view_path(collection_id).is_file()

    def _get_view(self, collection_id: int):
        if collection_id not in self._view_cache:
            if not self.view_exists(collection_id):
                raise IOError
            self._view_cache[collection_id] = HDF5View(self.view_path(collection_id))
        return self._view_cache[collection_id]

    def handle_request(self, collection_id: int, request: str, model: Dict[str, Any]) -> Dict[str, Any]:
        """
        Handles REST requests related to views. This function implements the GET endpoint
        /collections/[collection_id]/view/[request]

        Parameters
        ----------
        collection_id: int
            Collection id corresponding to a view.
        request: str
            Requested data. Allowed options and corresponding DatasetView methods:
            - list: list_values
            - value: get_values
            - molecule: get_molecules
            - entry: get_entries
        model:
            REST model containing input options.

        Returns
        -------
        Dict[str, Any]:
            Dictionary corresponding to requested REST model
        """
        meta = {"errors": [], "success": False, "error_description": False, "msgpacked_cols": []}

        try:
            view = self._get_view(collection_id)
        except IOError:
            meta["success"] = False
            meta["error_description"] = f"View not available for collection #{collection_id}"
            return {"meta": meta, "data": None}

        if request == "entry":
            try:
                df = view.get_entries(subset=model["subset"])
            except KeyError:
                meta["success"] = False
                meta["error_description"] = "Unable to find requested entry."
                return {"meta": meta, "data": None}
        elif request == "molecule":
            series = view.get_molecules(model["indexes"], keep_serialized=True)
            df = pd.DataFrame({"molecule": series})
            df.reset_index(inplace=True)
            meta["msgpacked_cols"].append("molecule")
        elif request == "value":
            df, units = view.get_values(model["queries"], subset=model["subset"])
            df.reset_index(inplace=True)
        elif request == "list":
            df = view.list_values()
        else:
            meta["success"] = False
            meta["error_description"] = f"Unknown view request: {request}."
            return {"meta": meta, "data": None}

        # msgpack columns not supported by pyarrow
        pack_columns = []
        for col in df.columns:
            if len(df) > 0:
                sample = df[col].iloc[0]
                if isinstance(sample, np.ndarray):
                    pack_columns.append(col)
                elif isinstance(sample, list):
                    pack_columns.append(col)
                # Add any other datatypes that need to be handled specially go here

        for col in pack_columns:
            df[col] = df[col].apply(lambda x: serialize(x, "msgpack-ext"))
        meta["msgpacked_cols"] += pack_columns

        # serialize
        f = io.BytesIO()
        df.to_feather(f)
        df_feather = f.getvalue()

        if request == "value":
            data = {"values": df_feather, "units": units}
        else:
            data = df_feather

        meta["success"] = True

        return {"meta": meta, "data": data}



"""Msgpack Remaining Phase 2

Revision ID: e32b61e2516f
Revises: da7c6f141bcb
Create Date: 2019-08-12 10:13:09.694643

"""
from alembic import op
import sqlalchemy as sa

import os
import sys

sys.path.insert(1, os.path.dirname(os.path.abspath(__file__)))
from migration_helpers import msgpack_migrations
from qcelemental.util import msgpackext_dumps, msgpackext_loads

# revision identifiers, used by Alembic.
revision = "e32b61e2516f"
down_revision = "da7c6f141bcb"
branch_labels = None
depends_on = None


def upgrade():
    ## Task Queue
    table_name = "task_queue"
    update_columns = {"spec"}

    nullable = set()
    msgpack_migrations.json_to_msgpack_table_altercolumns(table_name, update_columns, nullable_true=nullable)

    ## Service Queue
    table_name = "service_queue"
    update_columns = {"extra"}

    nullable = set()
    msgpack_migrations.json_to_msgpack_table_altercolumns(table_name, update_columns, nullable_true=nullable)


def downgrade():
    raise ValueError("Cannot downgrade json to msgpack conversions")



"""Add queue_manager id to Base_results

Revision ID: 4bb79efa9855
Revises: 26cfd7b0439e
Create Date: 2019-07-26 15:31:14.811337

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = "4bb79efa9855"
down_revision = "26cfd7b0439e"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column("base_result", sa.Column("manager_name", sa.String(), nullable=True))
    op.create_foreign_key(None, "base_result", "queue_manager", ["manager_name"], ["name"], ondelete="SET NULL")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_constraint(None, "base_result", type_="foreignkey")
    op.drop_column("base_result", "manager_name")
    # ### end Alembic commands ###
